# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import json

from tests import *
from unify.apimanager import ApiManager


class ApiManagerDefaultOrg(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.api_manger = ApiManager(
            cluster=cluster_name,
            props=props,
            org_id=test_org
        )

        cls.test_here = sources.static_file_upload(
            name=str(uuid.uuid4()),
            org_id=test_org,
            content="tests/data_test.csv"
        )

    def test_delete_source(self):
        delete = self.api_manger.delete_source(
            source_id=self.test_here["id"]
        )

        self.assertTrue("message" in delete)
        self.assertEqual(delete["message"], "Success", delete)

    def test_graphs_list(self):
        graph_list = self.api_manger.graphs_list()
        self.assertTrue(len(graph_list) is 0, graph_list)

    def test_regular_pipeline_duplicate(self):
        dupli = self.api_manger.regular_pipeline_duplicate(
            pipeline_id=test_pipeline["pipeline"]["id"],
            new_name=str(uuid.uuid4())
        )
        self.assertTrue("id" in dupli, dupli)

    def test_export_source(self):
        export = json.loads(self.api_manger.export_source(
            dataset_ids=[test_dataset["id"]]
        ))
        self.assertTrue(len(export) > 0, export)
        self.assertTrue("file_content" in export[0], export[0])

    def test_dataset_list(self):
        export = self.api_manger.dataset_list()
        self.assertTrue(len(export) > 0, export)

    def test_pipeline_list(self):
        export = self.api_manger.pipeline_list()
        self.assertTrue(len(export) > 0, export)

    def test_create_pipeline_export_data(self):
        export = self.api_manger.create_pipeline_export_data(
            pipeline_id=test_pipeline["pipeline"]["id"],
            skip=[]
        )

        self.assertTrue("map_attributes" in export, export)
        self.assertTrue("functions" in export, export)
        self.assertTrue("pipeline" in export, export)
        self.assertTrue("sources" in export, export)
        self.assertTrue("templates" in export, export)

    def test_proceses_importing_pipeline_file(self):
        export = self.api_manger.create_pipeline_export_data(
            pipeline_id=test_pipeline["pipeline"]["id"],
            skip=[]
        )
        import_results = json.loads(self.api_manger.proceses_importing_pipeline_file(
            content=export,
            skip=[]
        ))

        self.assertTrue("pipeline_id" in import_results, import_results)
        self.assertEqual(import_results["org_id"], test_org, import_results)
        self.assertTrue("url" in import_results, import_results)
        self.assertTrue("warnings" in import_results, import_results)

    def test_get_all_hierarchies_display(self):
        export = self.api_manger.get_all_hierarchies_display()

        self.assertTrue(len(export) > 0, export)

    def test_get_single_hierarchy(self):
        export = self.api_manger.get_all_hierarchies()

        self.assertTrue(len(export) > 0, export)

    def test_create_hierarchy(self):
        export = self.api_manger.create_hierarchy(name=str(uuid.uuid4()))

        self.assertTrue("id" in export, export)

    def test_export_hierarchy(self):
        export = json.loads(self.api_manger.export_hierarchy(hierarchy=test_hierarchy["id"]))

        self.assertEqual(export["name"], test_hierarchy["name"], export)
        self.assertTrue("levels" in export)

    def test_import_hierarchy(self):
        fi = json.loads(self.api_manger.export_hierarchy(hierarchy=test_hierarchy["id"]))

        fi["name"] = str(uuid.uuid4())

        import_results = self.api_manger.import_hierarchy(
            content=json.dumps(fi)
        )

        self.assertTrue("id" in import_results, import_results)



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains methods to interact with Unify datasets api
"""
import json
import time
import uuid
import os
from tempfile import mkstemp

from unify.properties import Properties, ClusterSetting
from .apirequestsmng import ApiRequestManager
from .generalutils import csv_to_json
from .generalutils import json_to_csv
from .generalutils import create_schema_dataset
from .generalutils import stream_iterable
from .WaitingLibrary import Wait
from unify.helpers.SingleOrg import single_org
from unify.helpers.graph_ql.dataset_gql import DatasetGrapql


class Sources(ApiRequestManager):
    """
    Class to interact with dataset endpoints
    """

    def __init__(self, cluster=None, props=Properties(ClusterSetting.KEY_RING), org_id=None):
        """
        Class constructor

        :param cluster: Hostname
        :type cluster: str
        :param props: Instantiated Properties class
        :type props: class:`unify.properties.Properties`
        """

        super().__init__(cluster=cluster, props=props, org_id=org_id)

        try:

            self.epoch_time = int(time.time())
            self.sources_url = self.props.get_remote(self.cluster) + "io/v1/org/{}/sources"

            self.agents_url = self.props.get_remote(self.cluster) + "agents/v2/"

            self.piwebapi_test_url = self.agents_url + "piwebapi/org/{}/test"

            self.piwebapi_create_url = self.agents_url + \
                                       "piwebapi/org/{}/create?modeldata_only=true"

            self.piconfig_upload_url = self.agents_url + "piconfig/org/{}/model/?name=&serverName="

            self.piconfig_upload_url_no_params = self.agents_url + "piconfig/org/{}/model/"

            self.staticfile_upload_url = self.agents_url + "staticfile/org/{}/model"

            self.delete_source_url = self.agents_url + "piwebapi/org/{}/source/{}"

            self.stage_file_url = self.props.get_remote(self.cluster) + "datasets/v1/stage/file"

            self.post_data_set_schema_url = self.props.get_remote(
                self.cluster) + "datasets/v1/dataset"

            self.commit_data_set_url = self.props.get_remote(
                self.cluster) + "datasets/v1/dataset/{}/commit"

            self.labeling_sources = self.props.get_remote(
                self.cluster) + "datasets/v1/labeling"

            self.labeling_sources_2 = self.props.get_remote(
                self.cluster) + "datasets/v1/labeling/query"

            self.labeling_sources_facets_url = self.props.get_remote(
                self.cluster) + "datasets/v1/labeling/facets"

            self.append_url = self.props.get_remote(
                self.cluster) + "agents/v2/staticfile/org/{}/model/{}/append"

            self.download_dataset_url = self.props.get_remote(
                self.cluster) + "tags/org/{}/datasets/{}/download"

            self.get_commit = self.commit_data_set_url + "/{}"

            self.preview_source_url = self.props.get_remote(
                self.cluster) + "tags/org/{}/datasets/{}/preview"

            self.dataset_info_url = self.props.get_remote(
                self.cluster) + "tags/org/{}/pipelines/v2/datasets"

            self.gql_builder = DatasetGrapql()

            # Datastreams
            self.datastream_events_info_url = self.props.get_remote(
                self.cluster) + "datasets/v1/dataset/{}/events"

            self.datastream_download_events = self.props.get_remote(
                self.cluster) + "datasets/v1/dataset/{}/events/download"

        except Exception as error:
            raise error

    def get_status(self, org_id, dataset_id, commit_id):
        """
        Retrieves the status of the given commit id

        :param org_id: Org where the commit has occurred
        :type org_id: int or str
        :param dataset_id: Datasets id of whom the commit belongs to
        :type dataset_id: str
        :param commit_id: Commit id to be retrieved
        :type commit_id: str
        :return:
        """
        return self.get_commit_status(org_id=org_id, data_set_id=dataset_id, commit_id=commit_id)[0]

    def accert_status(self, org_id, dataset_id, commit_id, expected):
        """
        Accerts that the status of a given commit is what is expected

        :param org_id: Org where the commit has occurred
        :type org_id: int or str
        :param dataset_id: Dataset id of whom the commit belongs to
        :type dataset_id: str
        :param commit_id: Commit id to be retrieved
        :type commit_id: str
        :param expected: Expected status
        :type expected: str
        :return:
        """
        status = self.get_status(org_id=org_id, dataset_id=dataset_id, commit_id=commit_id)

        if "status" in status:

            tipe = status["status"]

            if "$type" in tipe:
                return expected in tipe["$type"]

        return False

    def is_commit_completed(self, org_id, dataset_id, commit_id):
        """
        Verifies that the status of the commit is completed

        :param org_id: Org where the commit has occurred
        :type org_id: int or str
        :param dataset_id: Dataset id of whom the commit belongs to
        :type dataset_id: str
        :param commit_id: Commit id to be asserted
        :type commit_id: str
        :return:
        """
        return self.accert_status(
            org_id=org_id,
            dataset_id=dataset_id,
            commit_id=commit_id,
            expected="Completed"
        )

    def divide_dataset_in(
            self, name, org_id, content, format="csv",
            convert_to_parquet="false", encoding='UTF-8', chunks=10000):
        """
        Upload a dataset through the static route. This method should be used when
        uploading a big file. It will split the file into
        smaller chunks and upload them sequentially.

        :param name: Dataset name to be cerated
        :type name: str
        :param org_id: Org where the dataset will be created
        :type org_id: int or str
        :param content: Dataset content
        :type content: str
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert file to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: File encoding
        :type encoding: str, optional
        :param chunks: Number of chunks to split the file
        :type chunks: int
        :return:
        """

        data = content.encode(encoding)

        jdata = csv_to_json(data)

        dataset_data = {}
        index = 0
        for aux_file in stream_iterable(container=jdata, chunk=chunks):
            aux_name = "{}_{}".format(name, index)

            file_dir, path = mkstemp(suffix=".csv")

            open(path, "w+").write(json_to_csv(aux_file))

            dataset_data[aux_name] = self.create_api_data_set(
                name=aux_name,
                org_id=org_id,
                file_path=path,
                format=format,
                convert_to_parquet=convert_to_parquet,
                encoding=encoding
            )

            Wait().until(
                self.is_commit_completed,
                "commit {} never completed".format(dataset_data[aux_name]["data_set_id"]),
                org_id,
                dataset_data[aux_name]["data_set_id"],
                dataset_data[aux_name]["commit_id"]
            )

            os.close(file_dir)
            index += 1

        return dataset_data

    @single_org
    def upload_big_dataset(
            self, name, org_id=None, *, content, format="csv",
            convert_to_parquet="false", encoding='UTF-8', chunks=10000):
        """
        Upload a dataset through the static route. This method should be used when
        uploading a big file. It will split the file into
        smaller chunks and upload them sequentially.

        :param name: Dataset name to be cerated
        :type name: str
        :param org_id: Org where the dataset will be created
        :type org_id: int or str
        :param content: Dataset content
        :type content: str
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert file to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: File encoding
        :type encoding: str, optional
        :param chunks: Number of chunks to split the file
        :type chunks: int
        :return:
        """

        data = content.encode(encoding)

        jdata = csv_to_json(data, encoding=encoding)

        first = True

        dataset_data = {
            "create": {},
            "append": []
        }
        for aux_file in stream_iterable(container=jdata, chunk=chunks):

            if first:

                first = False

                file_dir, path = mkstemp(suffix=".csv")

                open(path, "w", encoding=encoding).write(json_to_csv(aux_file))

                dataset_data["create"] = self.create_api_data_set(
                    name=name,
                    org_id=org_id,
                    file_path=path,
                    format=format,
                    convert_to_parquet=convert_to_parquet,
                    encoding=encoding
                )

                Wait().until(
                    self.is_commit_completed,
                    "commit {} never completed".format(dataset_data["create"]["data_set_id"]),
                    org_id,
                    dataset_data["create"]["data_set_id"],
                    dataset_data["create"]["commit_id"]
                )

                os.close(file_dir)

            else:

                file_dir, path2 = mkstemp(suffix=".csv")

                open(path2, "w", encoding=encoding).write(json_to_csv(aux_file))

                try:
                    added_data = self.add_data_to_existing_source(
                        name="{} {}".format(name, str(uuid.uuid4())[:4]),
                        org_id=org_id,
                        data_set_id=dataset_data["create"]["data_set_id"],
                        file_path=path2,
                        group_id=dataset_data["create"]["group_id"]
                    )

                    Wait().until(
                        self.is_commit_completed,
                        "commit {} never completed".format(dataset_data["create"]["data_set_id"]),
                        org_id,
                        dataset_data["create"]["data_set_id"],
                        added_data["commit_id"]
                    )

                    dataset_data["append"].append(added_data)

                    os.close(file_dir)

                except Exception as err:
                    print("skipping {}".format(repr(err)))

        return dataset_data

    @single_org
    def create_api_data_set(
            self,
            name,
            org_id=None,
            *,
            file_path,
            format="csv",
            convert_to_parquet="false",
            encoding="UTF-8"
    ):

        """
        Creates a dataset through the static file route.

        :param name: Name to of the new dataset
        :type name: str
        :param org_id: Org id where the dataset will be stored
        :type org_id: int or str
        :param file_path: Directory file path where the dataset contents are
        :type file_path: str
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: Source file encoding
        :type encoding: str, optional
        :return:
        """

        if file_path is None:
            raise Exception("File Path must not be None")

        content = open(file_path, "rb").read().decode(encoding)

        return self.create_api_data_set_with_content(
            name=name,
            org_id=org_id,
            content=content,
            format=format,
            convert_to_parquet=convert_to_parquet,
            encoding=encoding
        )

    @single_org
    def create_api_data_set_with_content(
            self,
            name,
            org_id=None,
            *,
            content,
            format="csv",
            convert_to_parquet="false",
            encoding="UTF-8"):
        """
        Creates a dataset through the static file route.

        :param name: Name of the new dataset
        :type name: str
        :param org_id: Org id where the dataset will be stored
        :type org_id: int or str
        :param content: dataset content in csv or tsv format
        :type content: str
        :param format: Content format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: Content encoding
        :type encoding: str, optional
        :return:
        """
        results = {}

        stage_result = self.stage_file_with_content(
            name=name,
            org_id=org_id,
            file_data=content,
            format=format,
            convert_to_parquet=convert_to_parquet,
            encoding=encoding
        )

        results["group_id"] = stage_result

        schema = create_schema_dataset(
            csv_data=content,
            name=name
        )
        initial_commit = self.post_file_schema(
            org_id=org_id,
            schema_data=json.dumps(schema),
        )

        results["data_set_id"] = initial_commit["id"]
        results["commit_id"] = initial_commit["commitId"]

        second_commit = self.append_command(
            org_id=org_id,
            data_set_id=results["data_set_id"],
            name=name,
            group_id=results["group_id"]
        )

        results.update(second_commit)

        return results

    @single_org
    def append_all(self, name, org_id=None, *, file_path, format="csv",
                   convert_to_parquet="false", encoding='UTF-8', chunks=10000, existing_dataset_id=None):
        """

        This method will split a data-set into chunks, and uplaod chunk by chunk. This is usefull for
        limiting and not overwelm the unify cluster. Use this method when trying to upload big datasets.

        :param name: Dataset name to be cerated
        :type name: str
        :param org_id: Org where the dataset will be created
        :type org_id: int or str
        :param file_path: Dataset file directory route
        :type file_path: str
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert file to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: File encoding
        :type encoding: str, optional
        :param chunks: Number of chunks to split the file
        :type chunks: int
        :return:
        """

        content = open(file_path, "rb").read().decode(encoding)

        data = content.encode(encoding)

        jdata = csv_to_json(data, encoding=encoding)

        first = True if existing_dataset_id is None else False

        dataset_data = {
            "create": {},
            "append": []
        }

        all_groups = []

        any_stage = False

        dataset_id = None if existing_dataset_id is None else existing_dataset_id

        for aux_file in stream_iterable(container=jdata, chunk=chunks):

            if first:

                first = False

                file_dir, path = mkstemp(suffix=".csv")

                open(path, "w", encoding=encoding).write(json_to_csv(aux_file))

                dataset_data["create"] = self.create_api_data_set(
                    name=name,
                    org_id=org_id,
                    file_path=path,
                    format=format,
                    convert_to_parquet=convert_to_parquet,
                    encoding=encoding
                )

                dataset_id = dataset_data["create"]["data_set_id"]

                Wait().until(
                    self.is_commit_completed,
                    "commit {} never completed".format(dataset_data["create"]["data_set_id"]),
                    org_id,
                    dataset_data["create"]["data_set_id"],
                    dataset_data["create"]["commit_id"]
                )

                os.close(file_dir)

            else:

                try:
                    stage_name = "file_{}".format(str(uuid.uuid4())[:5])

                    if len(json_to_csv(aux_file)):
                        stage_result = self.stage_file_with_content(
                            name=stage_name,
                            org_id=org_id,
                            file_data=json_to_csv(aux_file),
                            format=format,
                            convert_to_parquet=convert_to_parquet,
                            encoding=encoding
                        )

                        print(stage_result.decode())

                        all_groups.append(
                            {
                                "groupId": stage_result.decode(),
                                "name": stage_name
                            }
                        )

                        any_stage = True

                except Exception as err:
                    print("skipping {}".format(repr(err)))

        if any_stage:
            dataset_data["append"] = self.append_all_command(
                org_id=org_id,
                files=all_groups,
                data_set_id=dataset_id
            )

        return dataset_data

    @single_org
    def get_commit_status(self, org_id=None, *, data_set_id, commit_id):
        """
        Retrieves the status of the given commit id.

        :param org_id: Org id where the commit has occurred
        :type org_id: int or str
        :param dataset_id: Datasets id of whom the commit belongs to
        :type dataset_id: str
        :param commit_id: Commit id to be retrieved
        :type commit_id: str
        :return:
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others={"'Content-Type'": "application/data"}
        )

        url = self.get_commit.format(data_set_id, commit_id)

        test_request = self.session.get(url, headers=header)

        return json.loads(test_request.content.decode('utf8')), test_request.status_code

    @single_org
    def download_dataset_content(self, org_id=None, *, dataset_id):
        """
        Downloads the content of a given dataset.

        :param org_id: Org id where the dataset exists
        :type org_id: int or str
        :param dataset_id: Dataset id to be retrieved
        :type dataset_id: str
        :return:
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )

        get_source_request = self.session.get(
            self.download_dataset_url.format(org_id, dataset_id),
            headers=header
        )

        if get_source_request.status_code == 200:
            return get_source_request.content

        raise Exception(repr(get_source_request.content))

    @single_org
    def create_export_dataset(self, org_id=None, *, dataset_ids):
        """
        Create the content needed to export a dataset.
        This is usually used when using the import dataset.

        :param org_id: Org id where the dataset exists
        :type org_id: int or str
        :param dataset_ids: List containing the dataset ids
        :type dataset_ids: list of str
        :return:
        """
        datasets = self.get_sources(org_id=org_id)

        id_to_type = {}

        for dataset in datasets:
            id_to_type[dataset["id"]["id"]] = dataset

        all_datasets = []
        for dataset_id in dataset_ids:

            if dataset_id not in id_to_type:
                continue

            info = id_to_type[dataset_id]

            source_type = "Upload"

            if "labels" in info:
                if "raw" in info["labels"]:
                    if "ean_source_type" in info["labels"]["raw"]:
                        source_type = info["labels"]["raw"]["ean_source_type"]

            get_source_request = self.download_dataset_content(org_id=org_id, dataset_id=dataset_id)

            result = {
                "component": None,
                "id": dataset_id,
                "schema": info["schema"],
                "name": id_to_type[dataset_id]["name"],
                "type": source_type,
                "file_content": csv_to_json(csv_data=get_source_request)
            }

            all_datasets.append(result)

        return json.dumps(all_datasets)

    @single_org
    def add_data_content_to_existing_source(
            self, name, org_id=None, *, content,
            data_set_id, group_id=None, format="csv"):
        """
        Helper function to append data to existing dataset

        :param name: Name of the current file, used for staging the file
        :type name: str
        :param org_id: Org id where the target dataset exists
        :type org_id: int or str
        :param content: content of append data
        :param data_set_id: Dataset id where the content will be appended
        :type data_set_id: int or str
        :param group_id: Group identification that the dataset belongs
        :type group_id: str
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :return:
        """
        results = {}

        stage = self.stage_file_with_content(
            name=name,
            org_id=org_id,
            file_data=content,
            group_id=group_id,
            format=format,
        )

        results["group_id"] = stage

        final_commit = self.append_command(
            org_id=org_id,
            data_set_id=data_set_id,
            name=name,
            group_id=results["group_id"]
        )

        results.update(final_commit)

        return results

    @single_org
    def add_data_to_existing_source(
            self, name, org_id=None, *, file_path,
            data_set_id, group_id=None, format="csv"):
        """
        Helper function to append data to existing dataset

        :param name: Name of the current file, used for staging the file
        :type name: str
        :param org_id: Org id where the target dataset exists
        :type org_id: int or str
        :param file_path: Directory file path that contains the data to be appended
        :type file_path: str
        :param data_set_id: Dataset id where the content will be appended
        :type data_set_id: int or str
        :param group_id: Group identification that the dataset belongs
        :type group_id: str
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :return:
        """
        results = {}

        second_stage = self.stage_data(
            name=name,
            org_id=org_id,
            file_path=file_path,
            group_id=group_id,
            format=format
        )

        results["group_id"] = second_stage

        final_commit = self.append_command(
            org_id=org_id,
            data_set_id=data_set_id,
            name=name,
            group_id=results["group_id"]
        )

        results.update(final_commit)

        return results

    @single_org
    def overwrite_dataset(
            self, org_id=None, *, data_set_id, file_path, group_id=None,
            format="csv", convert_to_parquet="false", encoding="utf-8", encode=True):

        """
        Overwrites the contents of the given dataset with new content.

        :param org_id: Org id where the target dataset exists
        :type org_id: int or str
        :param data_set_id: Dataset id that its content will be overwritten
        :type data_set_id: str
        :param file_path: Directory file path that contains the data to be appended
        :type file_path: str
        :param group_id: Group identification that the dataset belongs
        :type group_id: str, optional
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert file to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: Encoding of the file
        :type encoding: int or str
        :return:
        """
        response = {}

        response["truncate"] = self.truncate_data_set(
            org_id=org_id,
            data_set_id=data_set_id
        )

        name = "overwirte{}".format(int(time.time()))

        response["stage"] = self.stage_data(
            name=name,
            org_id=org_id,
            file_path=file_path,
            group_id=group_id,
            format=format,
            convert_to_parquet=convert_to_parquet,
            encoding=encoding,
            encode=encode
        )

        response["append"] = self.append_command(
            org_id=org_id,
            data_set_id=data_set_id,
            group_id=response["stage"],
            name=name
        )

        return response

    def _commit_dataset_command(self, org_id, data_set_id, command):
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others={"'Content-Type'": "application/data"}
        )

        headers = dict(header.items())
        request = self.session.post(
            self.commit_data_set_url.format(data_set_id),
            headers=headers,
            json=command
        )

        if request.status_code in [200, 202]:
            append = json.loads(request.content)

            Wait(200).until(
                self.is_commit_completed,
                "commit {} never completed".format(data_set_id),
                org_id,
                data_set_id,
                append["commit_id"]
            )

            return append
        else:
            raise Exception(repr(request.content))

    @single_org
    def truncate_data_set(self, org_id=None, *, data_set_id):
        """
        Truncates the dataset content.

        :param org_id: Org where the target dataset exists
        :type org_id: int or str
        :param data_set_id: Dataset id where which contents will be truncated
        :type data_set_id: str
        :return:
        """
        command = {"commands": [{"$type": "truncate", "cause": []}]}
        return self._commit_dataset_command(org_id, data_set_id, command)

    @single_org
    def append_dataset(self, org_id=None, *, data_set_id, content):
        """
        Appends data to existing dataset.

        :param org_id: Org where the target dataset exists
        :type org_id: int or str
        :param data_set_id: Dataset id where the content will be appended
        :type data_set_id: str
        :param content: Data content that will be added to the dataset
        :type content: str
        :return:
        """
        file_dir, path = mkstemp(suffix=".csv")

        open(path, "wb").write(content.encode())

        os.close(file_dir)

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )

        files = {'file': open(path, "rb")}

        post_upload_file = self.session.post(
            self.append_url.format(org_id, data_set_id),
            headers=headers,
            files=files
        )

        if post_upload_file.status_code in [200, 201, 202]:
            response = json.loads(post_upload_file.content)

            return response

        raise Exception(repr(post_upload_file.content))

    def label(self, org_id, data_set_id, labels):
        """
        Method to create dataset event object
        :param org_id: Org where the event will be applied
        :param data_set_id: dataset identification where the event would be applied
        :param labels: labels that will be added to this event
        :return:
        """
        command = {
            "commands":
                [
                    {
                        "$type": "text-label-group",
                        "key": "ean.facets",
                        "values": labels,
                        "cause": [],
                    }
                ]
        }
        return self._commit_dataset_command(org_id, data_set_id, command)

    def append_command(self, org_id, data_set_id, group_id, name):
        """
        Executes append command to a staged file on a group id.

        :param org_id: Org where the target dataset exists
        :type org_id: int or str
        :param data_set_id: Dataset id to be appended
        :type data_set_id: str
        :param group_id: Group identification that the dataset belongs
        :param name: Name of the staged file
        :type name: str
        :return:
        """
        command = {
            "commands":
                [
                    self.single_append_command(group_id=group_id, name=name)
                ]
        }
        return self._commit_dataset_command(org_id, data_set_id, command)

    def append_all_command(self, org_id, data_set_id, files: list):

        command = {
            "commands":
                [
                    {
                        "$type": "append-all",
                        "files": files,
                        "overwrite": False,
                        "cause": []
                    }
                ]
        }

        open("append_all.json", "w+").write(json.dumps(command))
        return self._commit_dataset_command(org_id, data_set_id, command)

    def stage_data(
            self, name, org_id, file_path, group_id=None, format="csv",
            convert_to_parquet="false", encoding='UTF-8', encode=True):
        """
        Stages data to a given org and group id.

        :param name: Name of the staged file
        :type name: str
        :param org_id: Org where the target dataset exists
        :type org_id: int or str
        :param file_path: Directory file path with files to be staged
        :type file_path: str
        :param group_id: Group identification that the dataset belongs
        :type group_id: str, optional
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: Dataset file encoding
        :type encoding: str, optional
        :param encode: Encode the dataset with given encoding
        :type encode: bool, optional
        :return:
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others={"'Content-Type'": "application/octet-stream"}
        )

        if group_id is None:
            payload_data = {
                "name": name,
                "convertToParquet": convert_to_parquet,
                "format": format,
                "charset": encoding
            }
        else:
            payload_data = {
                "name": name,
                "convertToParquet": convert_to_parquet,
                "format": format,
                "groupId": group_id,
                "charset": encoding
            }

        if encode:
            file_data = open(file_path, "r").read().encode(encoding)
        else:
            file_data = file_path.encode(encoding)

        test_request = self.session.post(
            self.stage_file_url,
            headers=header,
            data=file_data,
            params=payload_data
        )

        return test_request.content

    def stage_file(self, name, org_id, file_path, group_id=None, format="csv",
                   convert_to_parquet="false", encoding='UTF-8', encode=False):
        """
        Stage data from file to a given org and group id.

        :param name: Name of the staged file
        :type name: str
        :param org_id: Org where the target dataset exists
        :type org_id: int or str
        :param file_path: Directory file path with files to be staged
        :type file_path: str
        :param group_id: Group identification that the dataset belongs
        :type group_id: str, optional
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert to parquet
        :type convert_to_parquet: bool, optional
        :param encoding: Dataset file encoding
        :type encoding: str, optional
        :param encode: Encode the dataset with given encoding
        :type encode: bool, optional
        :return:
        """

        file_data = open(file_path, "r").read().encode(encoding)

        return self.stage_file_with_content(
            name=name,
            org_id=org_id,
            file_data=file_data,
            group_id=group_id,
            format=format,
            convert_to_parquet=convert_to_parquet,
            encoding=encoding,
            encode=encode
        )

    def stage_file_with_content(
            self, name, org_id, file_data, group_id=None,
            format="csv", convert_to_parquet="false", encoding='UTF-8', encode=False
    ):
        """
        Stage data from content to a given org and group id.

        :param name: Name of the staged file
        :type name: str
        :param org_id: Org where the target dataset exists
        :type org_id: int or str
        :param file_data: Content with data to be staged
        :type file_data: str
        :param group_id: Group identification that the dataset belongs
        :type group_id: str, optional
        :param format: File format. Accepts "CSV" or "TSV"
        :type format: str, optional
        :param convert_to_parquet: Flag to convert to parquet.
        :type convert_to_parquet: bool, optional
        :param encoding: File encoding
        :type encoding: str, optional
        :param encode: Encode the dataset with given encoding
        :type encode: bool, optional
        :return:
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others={"'Content-Type'": "application/octet-stream"}
        )

        if group_id is None:

            payload_data = {
                "name": name,
                "convertToParquet": convert_to_parquet,
                "format": format,
                "charset": encoding
            }
        else:
            payload_data = {
                "name": name,
                "convertToParquet": convert_to_parquet,
                "format": format,
                "groupId": group_id,
                "charset": encoding
            }

        test_request = self.session.post(
            self.stage_file_url,
            headers=header,
            data=file_data.encode(encoding),
            params=payload_data
        )

        return test_request.content

    def post_file_schema(self, org_id, schema_data, encoding='UTF-8'):
        """
        Posts the schema of a file to be staged.

        :param org_id: Org where the file is going to be staged
        :type org_id: int or str
        :param schema_data: File schema
        :type schema_data: dict
        :param encoding: File encoding
        :type encoding: str, optional
        :return:
        """

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others={"'Content-Type'": "application/data"}
        )

        schema_request = self.session.post(
            self.post_data_set_schema_url,
            headers=header,
            data=schema_data.encode(encoding)
        )

        if schema_request.status_code == 200:
            results = json.loads(schema_request.content)

            Wait(200).until(
                self.is_commit_completed,
                "commit {} never completed".format(results["id"]),
                org_id,
                results["id"],
                results["commitId"]
            )
            return results

        raise Exception(repr(schema_request.content))

    def pi_config_upload(self, name, server_name, file_path, org_id):
        """
        DEPRECATED - Use static_file_upload instead! This method may be removed
        in the future on a major version bump.

        Uploads PI-CONFIG dataset.

        :param name: Dataset name
        :type name: str
        :param server_name: PI-Config data archive server name
        :type server_name: str
        :param file_path: Directory file path with files to be uploaded
        :type file_path: str
        :param org_id: Org where the dataset will be created
        :type org_id: int or str
        :return:
        """

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )

        upload_file = {'file': open(file_path, "rb")}

        query_strings = {"name": str(name), "serverName": str(server_name)}

        post_upload_pitag = self.session.post(
            self.piconfig_upload_url_no_params.format(org_id),
            params=query_strings,
            headers=headers,
            files=upload_file
        )

        if post_upload_pitag.status_code == 200:
            return json.loads(post_upload_pitag.content)

        raise Exception(repr(post_upload_pitag.content))

    @single_org
    def static_file_upload(self, name, content, org_id=None, params=None, encoding=None):
        """
        :param name: Dataset name to be created
        :type name: str
        :param content: Dataset content stored in a variable
        :type content: str
        :param org_id: Org where the dataset will be created
        :type org_id: str
        :param params: Additional http parameters
        :return:
        """

        static_file_upload_url = self.staticfile_upload_url

        query_strings = {
            "name": str(name)
        }
        if encoding:
            query_strings["charset"] = encoding

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )

        files = {'file': open(content, "rb")}

        if params is not None:
            for tup in params:
                if len(tup) == 2:
                    query_strings[tup[0]] = tup[1]

        post_upload_file = self.session.post(
            static_file_upload_url.format(org_id),
            headers=headers,
            files=files,
            params=query_strings
        )

        if post_upload_file.status_code == 200:
            return json.loads(post_upload_file.content)

        raise Exception(repr(post_upload_file.content))

    @single_org
    def get_sources(self, org_id=None, page_num=1, page_size=200):

        """
        Retrieves all the metadata of datasets on an org.

        :param org_id: Org to be queried
        :type org_id: int or str
        :param page_num: pagination page number
        :param page_size: pagination page size
        :return:
        """

        query = self.gql_builder.build_dataset_query(
            page_num=page_num,
            page_size=page_size
        )

        get_sources_request = self.graph_ql_query(org_id=org_id, query=query)

        if get_sources_request.status_code in range(200, 203):
            data = json.loads(get_sources_request.content)

            if "data" in data:
                if "artifacts" in data["data"]:
                    return data["data"]["artifacts"]

            return []
        error_message = {
            "message": repr(get_sources_request.content),
            "code": get_sources_request.status_code
        }

        raise Exception(json.dumps(error_message))

    def get_sources_by_labels(self, org_id, facets):
        """
        Retieves all datasets that matches given labels.

        :param org_id: Org id
        :type org_id: int or str
        :param facets: List of dataset labels. Example: ["label1", "label2"]
        :type facets: list of str
        :return:
        """

        query = self.gql_builder.build_dataset_query(
            facets=facets
        )

        get_sources_request = self.graph_ql_query(org_id=org_id, query=query)

        if get_sources_request.status_code in range(200, 203):
            data = json.loads(get_sources_request.content)

            if "data" in data:
                if "artifacts" in data["data"]:
                    return data["data"]["artifacts"]

            return []

        raise Exception(repr(get_sources_request.content))

    def get_sources_by_filters(self, org_id, filters):
        """
        Retieves all datasets that matches given filters.

        :param org_id: Org id
        :type org_id: int or str
        :param filters: Dictonary with filter by name and value. Example: {'name':dataset_name}
        :type filters: dict of key value pair
        :return:
        """

        query = self.gql_builder.build_dataset_query(
            other_filters=filters
        )
        get_sources_request = self.graph_ql_query(org_id=org_id, query=query)

        if get_sources_request.status_code in range(200, 203):
            data = json.loads(get_sources_request.content)
            if "data" in data:
                if "artifacts" in data["data"]:
                    return data["data"]["artifacts"]

            return []

        raise Exception(repr(get_sources_request.content))

    @single_org
    def delete_source(self, org_id=None, *, source_id):
        """
        This method deletes a dataset that is stored on the org_id with the given source_id

        :param org_id: Organization ID where the datasets is stored
        :type org_id: int or str
        :param source_id: Data set id to be deleted
        :type source_id: str
        :return: Result of the delete operation {"message":"Success","metadata":{}}
        :raise: Exception when response code is not 200 or 201
        """

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        delete_request = self.session.delete(
            self.delete_source_url.format(
                str(org_id),
                source_id
            ),
            headers=headers
        )

        if delete_request.status_code in range(200, 203):
            return json.loads(delete_request.content)
        try:
            msg = json.loads(delete_request.content)
            raise Exception(msg)
        except json.JSONDecodeError as e:
            raise Exception(str(delete_request.content))

    @single_org
    def preview_dataset(self, org_id=None, *, dataset_id, page_num=1):

        """
        Preview the content of the dataset id give, page_num is the pagination of the given dataset
        :param org_id:
        :param dataset_id:
        :param page_num:
        :return:
        """

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        data = [("pageNum", page_num)]

        get_response = self.session.get(
            self.preview_source_url.format(org_id, dataset_id),
            headers=headers,
            params=data
        )

        if get_response.status_code in [200, 201]:
            return json.loads(get_response.content)

        try:
            msg = json.loads(get_response.content)
            raise Exception(msg)
        except json.JSONDecodeError:
            raise Exception(str(get_response.content))

    @single_org
    def get_dataset_info(self, org_id=None, *, dataset_id):

        """
        return metdadata of the desired dataset
        :param org_id:
        :param dataset_id:
        :return:
        """

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        data = [("dataset_id", dataset_id)]

        get_response = self.session.get(
            self.dataset_info_url.format(org_id),
            headers=headers,
            params=data
        )

        if get_response.status_code in [200, 201]:
            return json.loads(get_response.content)

        try:
            msg = json.loads(get_response.content)
            raise Exception(msg)
        except json.JSONDecodeError:
            raise Exception(str(get_response.content))

    @single_org
    def upload_spreadsheet(self, name=None, file_path=None, org_id=None, sheet_name=None):

        """
        :param name:
        :param file_path:
        :param org_id:
        :param sheet_name: Specify the spreadsheet name where the data is located
        :return:
        """
        if sheet_name is None:
            return self.static_file_upload(
                name=name,
                content=file_path,
                org_id=org_id
            )
        else:
            return self.static_file_upload(
                name=name,
                content=file_path,
                org_id=org_id,
                params=[('sheetName', sheet_name)]
            )

    @single_org
    def update_dataset_metadata(self, org_id=None, *, dataset_id=None, name=None, description=None, facets=None):
        """
        Method to update a dataset's metadata

        :param org_id: org id where the pipeline is stored
        :param dataset_id: id of the pipeline to be edited
        :param name: new pipeline name
        :param description: description to be added
        :param facets:
        :return:
        """

        return self.update_artifact_metadata(
            org_id=org_id,
            artifact_id=dataset_id,
            artifact_type="dataset",
            name=name,
            description=description,
            facets=facets
        )

    @single_org
    def wait_for_dataset(self, org_id=None, *, dataset_id):
        """
        Wait for a dataset to be ready to be used

        :param org_id: Organization id where the dataset was created
        :param dataset_id: dataset id number
        :return:
        """
        return self.wait_for_artifact(
            org_id=org_id,
            artifact_id=dataset_id,
            artifact_type="dataset"
        )

    @single_org
    def get_datastream_events_info(self, org_id=None, *, dataset_id, sequenceNr=0):
        """
        Returns a list with all events occurred in a datastream

        :param org_id: Optional, Organization id where the dataset was created
        :param dataset_id: dataset id number
        :param sequenceNr: Optional, the number of the first event to be returned
        :return: A list with the events
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )
        params = {'fromSequenceNr': sequenceNr}
        response = self.session.get(
            self.datastream_events_info_url.format(dataset_id),
            params=params,
            headers=header
        )

        if response.status_code in self.OK:
            result = response.json()
            return result['events']

        try:
            msg = json.loads(response.content)
            raise Exception(msg)
        except json.JSONDecodeError:
            raise Exception(str(response.content))

    @single_org
    def get_datastream_event_data(self, org_id=None, *, dataset_id, sequenceNr):
        """
        Returns the data of the datastream event specified in the 'sequenceNr' param

        :param org_id: Optional, Organization id where the dataset was created
        :param dataset_id: dataset id number
        :param sequenceNr: the number of the datastream event to be returned
        :return: A csv string with the data starting from event specified
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )
        params = {'fromSequenceNr': sequenceNr}
        response = self.session.get(
            self.datastream_download_events.format(dataset_id),
            params=params,
            headers=header
        )

        if response.status_code in self.OK:
            return response.text

        try:
            msg = json.loads(response.content)
            raise Exception(msg)
        except json.JSONDecodeError:
            raise Exception(str(response.content))

    @single_org
    def create_command(self, org_id=None, *,
                       schema: dict,
                       name: str,
                       facets: list,
                       description: str,
                       cause: list,
                       dataset_uuid=None,
                       is_stream: bool = False):
        """

        :param org_id: organization id
        :type org_id: str
        :param schema: dataset schema
        :type schema: dict
        :param name: datatset name
        :type name: str
        :param facets: labels to be include don the dataset
        :type facets: list
        :param description: Description of the dataset
        :type description: str
        :param cause: The reason for this command
        :type cause: list
        :param dataset_uuid: Override of the dataset id
        :type dataset_uuid: str
        :param is_stream: decides if the created dataset is a stream dataset
        :type is_stream: bool
        :return: server response from the server
        """

        body = {
            "commands": [
                self.single_create_command(
                    schema=schema,
                    org_id=org_id,
                    name=name,
                    is_stream=is_stream,
                    facets=facets,
                    description=description,
                    cause=cause)
            ]
        }

        data_set_id = dataset_uuid if dataset_uuid is not None else str(uuid.uuid4())

        try:

            response = self._commit_dataset_command(org_id=org_id, data_set_id=data_set_id, command=body)

            response["uuid"] = data_set_id

            return response

        except Exception as es:
            raise Exception(es)

    def single_append_command(self, group_id, name):

        return {"$type": "append", "name": name, "group": group_id.decode(), "cause": []}

    def single_create_command(self, schema, org_id, name, is_stream, facets, description, cause):
        return {
            "$type": "create",
            "schema": schema,
            "org": org_id,
            "name": name,
            "isBounded": not is_stream,
            "facets": facets,
            "description": description,
            "cause": cause
        }



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains cluster definition
"""

class Cluster:
    """
    Class to store cluster data on memory
    """

    def __init__(self, userName, password, cluster, token):
        """
        Constructor

        :param userName: Cluster login username
        :type userName: str
        :param password: Cluster login password
        :type password: str
        :param cluster: Cluster url, format http://host
        :type cluster: str
        :param token: Auth token
        :type token: str
        """
        self.userName = userName
        self.password = password
        self.cluster = cluster
        self.token = token

    def get_password(self):
        """
        Get cluster password

        :return:
        """
        return self.password

    def get_username(self):
        """
        Get cluster password

        :return:
        """
        return self.userName



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unify.graph import Graph
from unify.apimanager import ApiManager
from tests import *


class TestGraphs(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.graph = Graph(cluster_name, props)
        cls.api_manager = ApiManager(cluster=cluster_name, props=props)

    def test_retrieve_graph_list_via_manager(self):

        """
        Verify if sdk can retrieve graphs list using api manager
        :return:
        """
        raised = False
        try:
            self.api_manager.graphs_list(
                org_id=test_org
            )
        except Exception as e:
            raised = True

        self.assertFalse(raised, 'Exception raised in retrieving graphs list')

    def test_retrieve_graph_list(self):

        """
        Verify if sdk can retrieve graphs list using regular client
        :return:
        """
        raised = False
        try:
            self.graph.get_graphs_list(
                org_id=test_org
            )
        except Exception as e:
            raised = True

        self.assertFalse(raised, 'Exception raised in retrieving graphs list')

    def test_query_graph(self):
        """
        Verify if the graph endpoint can be reached by the sdk
        :return:
        """
        name = str(uuid.uuid4())
        try:
            _ = self.graph.query_graph(
                org_id=test_org,
                graph=name,
                query="MATCH (n)-[r]->(g) RETURN *"
            )

        except Exception as e:

            self.assertTrue(name in str(e), e)

    def test_create_empty_graph(self):
        """
        Verify empty graph endpoint is reachable

        :return:
        """
        name = str(uuid.uuid4())
        raised = False
        try:
            m = self.graph.create_empty_graph(
                org_id=test_org,
                graph_name=name
            )
        except Exception as e:
            m = str(e)
            raised = True

        self.assertFalse(raised, m)



class Verbs:
    modifyworkspace = "modifyworkspace"
    list = "list"
    read = "read"
    write = "write"
    anyverb = "anyverb"
    everyverb = "everyverb"


class Domains:
    template = "template"
    pipeline = "pipeline"
    dataset = "dataset"
    graph = "graph"
    connector = "connector"
    permission = "permission"
    org_api = "org_api"
    workspace = "workspace"


class Effects:
    allow = "allow"
    deny = "deny"


class DisplayNames:
    everyone = "everyone"
    nobody = "nobody"
    everything = "everything"
    nothing = "nothing"


class SelectorBool:
    everyone = 1
    nobody = 2
    everything = 3
    nothing = 4


class SelectorType:
    user = "user"
    resource = "resource"


class ArtifactType:
    connector = "connector"
    dataset = "dataset"
    pipeline = "pipeline"
    graph = "graph"
    template = "template"



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains helper waiting methods
"""
import time
POLL_FREQUENCY = 2  # How long to sleep in between
IGNORED_EXCEPTIONS = ()  # exceptions ignored during calls to the method
TIME_OUT = 10


class Wait:
    """
    Unify wait library
    """

    def __init__(self, timeout=TIME_OUT, poll_frequency=POLL_FREQUENCY, ignored_exceptions=None):
        """
        Unify wait library constuctor

        :param timeout: Max time to wait before timeout
        :type timeout: int, optional
        :param poll_frequency: How often the method should be verified
        :type poll_frequency: int, optional
        :param ignored_exceptions: Ignore exceptions
        :type ignored_exceptions: bool, optional
        """

        self._timeout = timeout
        self._poll = poll_frequency

        if self._poll == 0:
            self._poll = POLL_FREQUENCY

        exceptions = list(IGNORED_EXCEPTIONS)

        if ignored_exceptions is not None:

            try:
                exceptions.extend(iter(ignored_exceptions))

            except TypeError:

                exceptions.append(ignored_exceptions)

        self._ignored_exceptions = tuple(exceptions)

    def until(self, method, message='', *args):
        """
        Waits for method to return TRUE.

        :param method: Pointer to a python method
        :type method: function
        :param message: String message to return if method times out
        :type message: str, optional
        :param args: Additional args to be passed to method
        :type args: args
        :return:
        """

        end_time = time.time() + self._timeout
        while True:
            try:
                value = method(*args)
                if value:
                    return value
            except self._ignored_exceptions as exc:
                message += str(getattr(exc, 'stacktrace', None))
                break

            time.sleep(self._poll)
            if time.time() > end_time:
                break
        raise Exception(message)

    def until_not(self, method, message='', *args):
        """
        Negate until

        :param method: Pointer to a python method
        :type method: method
        :param message: String message to return if method times out
        :type message: str, optional
        :param args: Additional args to be passed to method
        :type args: args
        :return:
        """

        end_time = time.time() + self._timeout

        while True:
            try:
                value = method(*args)
                if not value:
                    return value
            except self._ignored_exceptions as exc:
                message += str(getattr(exc, 'stacktrace', None))
                break

            time.sleep(self._poll)

            if time.time() > end_time:
                break

        raise Exception(message)

    def until_backoff(self, method, message='', *args):
        """
        Waits until with exponential backoff.

        :param method: Pointer to a python method
        :type method: method
        :param message: String message to return if method times out
        :type message: str, optional
        :param args: Additional args to be passed to method
        :type args: args
        :return:
        """
        end_time = time.time() + self._timeout

        new_poll = self._poll

        while True:
            try:
                value = method(*args)
                if value:
                    return value
            except self._ignored_exceptions as exc:
                message += str(getattr(exc, 'stacktrace', None))
                break

            time.sleep(new_poll)

            new_poll *= 0.02

            if time.time() > end_time:
                break
        raise Exception(message)



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import uuid

from unify.orgadmin import OrgAdmin
from unify.properties import Properties
from unify.properties import ClusterSetting
from unify.sources import Sources
from unify.pipelines import Pipelines
from unify.apimanager import ApiManager
from unify.connectors import Connectors
from unify.WaitingLibrary import Wait

cluster_name = "data"


def user_is_loged_in_to_org(prp, org_id):
    with Sources(cluster=cluster_name, props=prp) as s_admin:
        try:
            s_admin.get_sources(org_id=org_id)
            return True
        except:
            return False


props = Properties(clusterSetting=ClusterSetting.MEMORY)

props.store_cluster(
    username=os.environ.get("username"),
    password=os.environ.get("password"),
    name=cluster_name,
    cluster=os.environ.get("cluster")
)

org_admin = OrgAdmin(props=props, cluster=cluster_name)

test_org = org_admin.create_organization(
    org_name="zQA-{}".format(str(uuid.uuid4()))
)

Wait(timeout=200).until(user_is_loged_in_to_org, "User never available in org", props, test_org)

sources = Sources(cluster=cluster_name, props=props)

test_dataset = sources.static_file_upload(
    name="table_{}".format(str(uuid.uuid4())[:12]),
    org_id=test_org,
    content="tests/data_test.csv"
)


pipelines = Pipelines(cluster=cluster_name, props=props)

test_pipeline = pipelines.create_pipeline(name=str(uuid.uuid4()), org_id=test_org)

pipelines.wait_for_pipeline(
    org_id=test_org,
    pipeline_id=test_pipeline["pipeline"]["id"]
)

api_manager = ApiManager(cluster=cluster_name, props=props)

test_hierarchy = api_manager.create_hierarchy(name=str(uuid.uuid4()), org_id=test_org)

connectors = Connectors(cluster=cluster_name, props=props)

test_connector = None
# test_connector = connectors.create_connector(
#    org_id=test_org,
#    name=str(uuid.uuid4()),
#    connector_type='connector_test_type',
#    config={'a':'b'},
# )


print("--------Created organization {} for testing------", format(test_org))



import json


class Template:

    def __init__(self):
        self.attribute_name = ""
        self.template_id = ""
        self.sanitized_name = ""
        self.attribute_type = ""
        self.data_type = ""
        self.description = None
        self.interpolation = None
        self.uom = None

    def to_json(self):
        return json.dumps(self.build())

    def build(self):

        payload = {
            "name": self.attribute_name,
            "templateId": self.template_id,
            "sanitizedName": self.sanitized_name,
            "attributeType": self.attribute_type,
            "dataType": self.data_type
        }

        if self.description is not None:
            payload["description"] = self.description

        if self.interpolation is not None:
            payload["interpolationMethod"] = self.interpolation

        if self.uom is not None:
            payload["uom"] = self.uom

        return payload

    def __str__(self):

        return str(self.build())



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains methods to interact with graphs api
"""
import json
import io

import json_lines

from unify.properties import Properties
from unify.properties import ClusterSetting
from unify.apirequestsmng import ApiRequestManager
from unify.helpers.SingleOrg import single_org
from unify.helpers.graph_ql.permissions import PermissionsGraphQl
from unify.helpers.Permissions import DisplayNames


class Permissions(ApiRequestManager):
    """
    Class to interact with pipeline endpoints
    """

    def __init__(self, cluster=None, props=Properties(ClusterSetting.KEY_RING), org_id=None):
        """
        Class constructor

        :param cluster: Cluster name to be used
        :type cluster: str
        :param props: Properties instantiated object
        :type props: class:`unify.properties.Properties`
        """
        super().__init__(cluster=cluster, props=props, org_id=org_id)

        remote = self.props.get_remote(self.cluster)

        self.query_builder = PermissionsGraphQl()

    def get_org_config(self, org_id):
        """
        Retrieves the organization permissions configuration

        :param org_id: Organization ID to be queried
        :return:
        """
        resp = self.graph_ql_query(
            org_id=org_id,
            query=self.query_builder.get_org_config()
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def get_rules(self, org_id):
        """
        Retrieves the organization id applied rules

        :param org_id: Organization ID to be queried
        :return:
        """
        resp = self.graph_ql_query(
            org_id=org_id,
            query=self.query_builder.get_rules()
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def get_user_rules(self, org_id, user_id):
        """
        Retrieves the applied rules to a given user

        :param org_id: Organization id where the user is stored
        :param user_id: User ID to be queried
        :return:
        """

        resp = self.graph_ql_query(
            org_id=org_id,
            query=self.query_builder.get_user_rules(user_id=user_id)
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def get_artifact_rules(self, org_id, artifact_id, artifact_type):
        """
        Retrieves the applied rules to a given artifact

        :param org_id: Organization id where the artifact is stored
        :param artifact_id: Artifact Identification UUID
        :param artifact_type: Dataset or Pipeline
        :return:
        """
        query = self.query_builder.get_artifact_rules(
            artifact_id=artifact_id,
            artifact_type=artifact_type
        )

        resp = self.graph_ql_query(
            org_id=org_id,
            query=query
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def get_selector_definitions(self, org_id):
        """
        Retrieve selector definitions of a given org

        :param org_id: Organization ID
        :return:
        """
        resp = self.graph_ql_query(
            org_id=org_id,
            query=self.query_builder.get_selector_definitions()
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def check_permission(self, org_id, artifact_type, artifact_id, user_id, verb):
        """
        Checks the permissions applied to a given artifact and user

        :param org_id: Organization identification where the artifact and user are stored
        :param artifact_type: Dataset or Pipeline
        :param artifact_id: Artifact Identification UUID
        :param user_id: User identification UUID
        :param verb:
        :return:
        """
        query = self.query_builder.check(
            artifact_type=artifact_type,
            artifact_id=artifact_id,
            user_id=user_id,
            verb=verb
        )

        resp = self.graph_ql_query(
            org_id=org_id,
            query=query
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def get_selector_definition(self, org_id, selector_id):
        """
        Retrieve the selector definition

        :param org_id:  Organization identification where the selector is stored
        :param selector_id: Selector ID to query
        :return:
        """

        resp = self.graph_ql_query(
            org_id=org_id,
            query=self.query_builder.get_selector_definition(selector_id=selector_id)
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def add_rule(self, org_id, domain, verb, effect, userSelector, resourceSelector):
        """
        Adds a new permission rule

        :param org_id:  Organization identification where the selector is stored
        :param domain: unify.helpers.Permissions.Domains
        :param verb: unify.helpers.Permissions.Verbs
        :param effect: unify.helpers.Permissions.Effects
        :param userSelector: unify.helpers.Permissions.SelectorBool
        :param resourceSelector: unify.helpers.Permissions.SelectorBool
        :return:
        """

        query = self.query_builder.build_rule_mutation(
            domain=domain,
            verb=verb,
            effect=effect,
            userSelector=userSelector,
            resourceSelector=resourceSelector
        )

        resp = self.graph_ql_query(
            org_id=org_id,
            query=query
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))

    def get_specific_selectors(self, org_id, user: bool, display_name: str = None):
        """
        Retrieve a given selector

        :param org_id: Organization identification where the selector is stored
        :param user: User identification UUID
        :param display_name: The selector name
        :return:
        """
        selectors = self.get_selector_definitions(org_id=org_id)

        selector_type = "user" if user else "resource"

        selected = []

        if "data" in selectors:

            data = selectors["data"]

            if "permissions" in data:

                per = data["permissions"]

                if "getSelectorDefinitions" in per:

                    for selector in per["getSelectorDefinitions"]:

                        if selector["selectorType"] == selector_type:
                            if display_name:
                                if selector["displayName"] == display_name:
                                    selected.append(selector["id"])
                            else:
                                selected.append(selector["id"])

        return selected

    def get_everyone_user_selectors(self, org_id):
        return self.get_specific_selectors(org_id=org_id, user=True, display_name=DisplayNames.everyone)

    def get_nobody_user_selectors(self, org_id):
        return self.get_specific_selectors(org_id=org_id, user=True, display_name=DisplayNames.nobody)

    def get_everything_resource_selectors(self, org_id):
        return self.get_specific_selectors(org_id=org_id, user=False, display_name=DisplayNames.everything)

    def get_nothing_resource_selectors(self, org_id):
        return self.get_specific_selectors(org_id=org_id, user=False, display_name=DisplayNames.nothing)

    def delete_rule(self, org_id, rule_id):
        resp = self.graph_ql_query(
            org_id=org_id,
            query=self.query_builder.build_delete_rule_mutation(
                rule_id=rule_id
            )
        )

        if resp.status_code in self.OK:
            return json.loads(resp.content)

        raise Exception(repr(resp.content))



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains methods to interact with pipeline api
"""
import json
import uuid

from tempfile import mkstemp

from unify.generalutils import json_to_csv
from unify.properties import Properties
from unify.properties import ClusterSetting
from unify.apirequestsmng import ApiRequestManager
from unify.helpers.SingleOrg import single_org
from unify.helpers.graph_ql.pipeline_gql import PipelineGrapql


class Pipelines(ApiRequestManager):
    """
    Class to interact with pipeline endpoints
    """

    def __init__(self, cluster=None, props=Properties(ClusterSetting.KEY_RING), org_id=None):
        """
        Class constructor

        :param cluster: Cluster name to be used
        :type cluster: str
        :param props: Properties instantiated object
        :type props: class:`unify.properties.Properties`
        """

        super().__init__(cluster=cluster, props=props, org_id=org_id)

        try:

            remote = self.props.get_remote(self.cluster)

            self.pipelines_url = self.props.get_remote(
                self.cluster) + "tags/org/" + '{}' + "/pipelines"

            self.regular_duplicate_url = self.pipelines_url + "/v2/{}/duplicate"

            self.delete_pipeline_url = self.pipelines_url + '/{}'

            self.pipeline_url = self.pipelines_url + '/{}'

            self.flow_url = self.pipeline_url + '/flows/{}'

            self.run_pipeline_url = self.pipeline_url + '/run'

            self.run_pipeline_status_url = self.pipeline_url + '/status'

            self.create_map_attributes_transf = self.pipeline_url + "/components/{}/map-attribute-rules"

            self.pipelines_url_v2 = self.pipelines_url + "/v2/{}"

            self.autosync_url = remote + "tags/org/{}/pipelines/{}/autosync"

            self.retrieve_pipelines_v2 = remote + "tags/org/" + '{}' + "/pipelines/v2"

            self.duplicate_pipeline_url = self.pipelines_url_v2 + "/duplicate"

            self.published_pipelines_url = remote + 'api/orgs/{}/graphs/current/pipelines'

            self.preview_data_pipeline = self.flow_url + '/preview?pageNum=1&sortKey={}&sortDir=ASC'

            self.download_map_attribute_rules = remote + 'tags/org/{}/pipelines/{}/components/{}/map-attribute-rules/csv'

            self.flow_preview_url = self.flow_url + '/preview?pageNum=1'

            self.download_flow_url = self.props.get_remote(
                self.cluster
            ) + "tags/org/{}/pipelines/{}/flows/{}/download"

            self.rules_by_templates_url = remote + "tags/org/{}/pipelines/{}/components/{}/map-attribute-rules/rules-by-template"

            self.distinct_values_map_attributes = self.props.get_remote(
                self.cluster) + "tags/org/{}/pipelines/{}/flows/{}/distinctValuesWithCount"

            self.apply_map_attributes_url = self.props.get_remote(
                self.cluster) + "tags/org/{}/pipelines/{}/components/{}/map-attribute-rules"

            self.count_endpoint = self.props.get_remote(self.cluster) + "tags/org/{}/pipelines/{}/flows/{}/count"

            self.distinct_count_values_url = self.props.get_remote(
                self.cluster) + 'tags/org/{}/pipelines/{}/flows/{}/distinctValuesWithCount'

            self.distinct_values_url = self.props.get_remote(
                self.cluster) + 'tags/org/{}/pipelines/{}/flows/{}/distinctValues'

            self.group_components_url = self.pipeline_url + "/groupComponents"

            self.gql_builder = PipelineGrapql()

        except Exception as error:
            raise error

    @single_org
    def apply_map_attributes(self, org_id=None, pipeline_id=None, component=None, payload=None):
        """
        Applies map attribute rules to given component

        :param org_id: Org id where the pipeline exists
        :type org_id: int or str
        :param pipeline_id: Pipeline where the map attributes component exists
        :type pipeline_id: int or str
        :param component: Map attribute component id where the rules are applied
        :type component: int or str
        :param payload: Dict representing mappings
        :type payload: dict
        :return:
        """

        final_url = self.apply_map_attributes_url.format(org_id, pipeline_id, component)

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.delete_content_type_header
        )

        result = self.session.post(final_url, headers=header, data=json.dumps(payload))

        if result.status_code == 200:
            return json.loads(result.content)

        raise Exception(repr(result.content))

    @single_org
    def regular_duplicate(self, org_id=None, *, pipeline_id, new_name):
        """
        Will duplicate the given pipeline on the given org
        :param org_id:
        :param pipeline_id:
        :param new_name:
        :return:
        """
        if pipeline_id is None:
            raise Exception("Pipeline id cant be null")
        if new_name is None:
            raise Exception("New pipeline name cant be none")

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others={"Content-Type": "application/json"}
        )

        upload_url = self.regular_duplicate_url.format(org_id, pipeline_id)

        post_upload_file = self.session.post(
            upload_url,
            headers=headers,
            data=json.dumps({"name": new_name})
        )

        if post_upload_file.status_code == 200:
            response = json.loads(post_upload_file.content)

            if "sourceMetadata" in response:
                del response["sourceMetadata"]

            return response

        raise Exception(repr(post_upload_file.content))

    @single_org
    def update_pipeline_from_json(self, org_id=None, update_payload=None, pipeline_id=None, pipeline_name=None,
                                  sources=None):
        """
        Updates a pipeline with a json content

        :param org_id: Org id where the pipeline is located
        :type org_id: int or str
        :param update_payload: JSON payload that contains pipeline content
        :type update_payload: dict
        :param pipeline_id: Pipeline identification
        :type pipeline_id: int or str
        :param pipeline_name: Pipeline current name
        :type pipeline_name: str
        :param sources: List of sources id to use
        :type sources: list of str
        :return:
        """

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        update_payload["name"] = str(pipeline_name)

        index = 0

        for component in update_payload["components"]:
            if sources is not None and len(sources) > 0:
                if component["jsonClass"] == "JsonSourceRef":

                    if self.evergreen_enabled is False:
                        if "version" in component:
                            del component["version"]
                            component["sourceId"] = "100"

                    if isinstance(sources[index], int):
                        component["sourceId"] = int(sources[index])
                    else:
                        component["sourceId"] = sources[index]
                        if self.evergreen_enabled:
                            component["version"] = 1

                    index += 1
            if sources is None:
                if component["jsonClass"] == "JsonSourceRef":
                    if self.evergreen_enabled is False:
                        if "version" in component:
                            del component["version"]
                            component["sourceId"] = 1000

            if self.evergreen_enabled:
                if "version" not in component:
                    if component["jsonClass"] == "JsonSourceRef":
                        component["version"] = 1
                        component["sourceId"] = str(uuid.uuid4())

            if component["jsonClass"] == "JsonGraphSink":
                component["orgId"] = int(org_id)

        update_pipeline_request = self.session.put(
            self.pipeline_url.format(
                org_id, pipeline_id
            ),
            headers=header,
            data=json.dumps(update_payload)
        )

        if update_pipeline_request.status_code == 200:
            return json.loads(update_pipeline_request.content)

        raise Exception(repr(update_pipeline_request.content))

    @single_org
    def download_map_attributes(self, org_id=None, *, pipeline_id, component_id):
        """
        Downloads the map attributes data from the given pipeline

        :param org_id: Org id where the pipleine exists
        :type org_id: int or str
        :param pipeline_id: Pipeline id where the componentn exists
        :type pipeline_id: int or str
        :param component_id: Component where the map attributes are located
        :type component_id: int or str
        :return:
        """
        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        download_rules = self.download_map_attribute_rules.format(
            org_id,
            pipeline_id,
            component_id
        )

        result = self.session.get(download_rules, headers=headers)

        if result.status_code == 200:
            return result.content

        raise Exception(repr(result.content))

    @single_org
    def upload_map_attributes_from_json(self, org_id=None, pipeline_id=None, component_id=None, json_data=None):
        """
        Uploads a json file to the given map attributes component

        :param org_id: Org id where the pipeline exists
        :type org_id: int or str
        :param pipeline_id: Pipeline id where to component id exists
        :type pipeline_id: int or str
        :param component_id: Map attributes component id exists
        :type component_id: int or str
        :param json_data: Contains the map attribute rules
        :type json_data: dict
        :return:
        """
        _, path = mkstemp(suffix=".csv")

        open(path, "wb").write(json_to_csv(data_array=json_data).encode())

        self.upload_map_attributes(
            org_id=org_id,
            pipeline_id=pipeline_id,
            component_id=component_id,
            csv_file=path
        )

    @single_org
    def upload_map_attributes(self, org_id=None, pipeline_id=None, component_id=None, csv_file=None):
        """
        Uploads a csv file to the given map attributes component

        :param org_id: Org id where the pipeline exists
        :type org_id: int or str
        :param pipeline_id: Pipeline id where to component id exists
        :type pipeline_id: int or str
        :param component_id: Map attributes component id exists
        :type component_id: int or str
        :param csv_file: CSV file that contains the map attribute rules
        :type csv_file: str
        :return:
        """

        headers = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others={"'Content-Type'": "application/data"}
        )

        files = {'file': open(csv_file, "rb")}

        upload_url = self.download_map_attribute_rules.format(org_id, pipeline_id, component_id)

        post_upload_file = self.session.post(
            upload_url,
            headers=headers,
            files=files
        )

        if post_upload_file.status_code == 200:
            response = json.loads(post_upload_file.content)

            if "sourceMetadata" in response:
                del response["sourceMetadata"]

            return response

        raise Exception(repr(post_upload_file.content))

    @single_org
    def create_pipeline(self, name, org_id=None, function=False):
        """
        Creates an empty pipeline

        :param name: New pipeline name
        :type name: str
        :param org_id: Org id where the pipeline is going to be created
        :type org_id: int or str
        :param function: Defines if the pipeline is a function or not
        :type function: boolean
        :return:
        """

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        payload = {"name": name, "components": []}

        if function:
            payload["pipelineType"] = "function"

        result = self.session.post(
            self.pipelines_url.format(org_id),
            headers=header,
            data=json.dumps(payload)
        )

        if result.status_code == 200:
            return json.loads(result.content)

        raise Exception(json.loads(result.content))

    @single_org
    def download_flow(self, org_id=None, pipeline_id=None, flow_id=None):
        """
        Downloads the flow data

        :param org_id: Org where the piprline exists
        :type org_id: int or str
        :param pipeline_id: Pipeline id where the flow exists
        :type pipeline_id: int or str
        :param flow_id: Flow id to be queried
        :type flow_id: int or str
        :return:
        """

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        cookies = {'_gid': 'GA1.1.1649742346.1580648966'}

        schema_url = self.download_flow_url.format(org_id, pipeline_id, flow_id)

        result = self.session.get(schema_url, headers=header, cookies=cookies)

        if result.status_code == 200:
            return result.content

        raise Exception(repr(result.content))

    @single_org
    def get_pipeline(self, org_id=None, *, pipeline_id):
        """
        Retrieves pipeline json blob

        :param org_id: Org where the pipeline to be queried exists
        :type org_id: int or str
        :param pipeline_id: Pipeline id to be queried
        :type pipeline_id: int or str
        :return:
        """

        header = self.build_header(
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        pipeline_url = self.pipeline_url.format(org_id, pipeline_id)

        result = self.session.get(pipeline_url, headers=header)

        if result.status_code == 200:
            return json.loads(result.content)

        raise Exception(repr(result.content))

    @single_org
    def delete_pipeline(self, org_id=None, *, pipeline_id):
        """
        Deletes the given pipeline
        :param org_id:
        :param pipeline_id:
        :return:
        """

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        result = self.session.delete(
            self.pipelines_url_v2.format(org_id, pipeline_id),
            headers=header
        )

        if result.status_code in [200, 201, 202]:
            return json.loads(result.content)

        raise Exception(repr(result.content))

    @single_org
    def unpublish_pipeline(self, org_id=None, *, pipeline_id):
        """
        Unpublishes a pipeline from graph v1
        :param org_id:
        :param pipeline_id:
        :return:
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        update_payload = {"status": "draft"}

        response = self.session.put(
            self.pipelines_url_v2.format(org_id, pipeline_id),
            headers=header,
            data=json.dumps(update_payload)
        )

        if response.status_code in [200, 201, 202]:
            return json.loads(response.content)

        raise Exception(repr(response.content))

    @single_org
    def get_map_attribute_rules(self, org_id=None, *, pipeline_id, component):
        """
        Downloads map attributes rules of the given pipeline component

        :param org_id: Org where pipeline lives
        :type org_id: int or str
        :param pipeline_id: Pipeline id to be queried
        :type pipeline_id: int or str
        :param component: Map attributes id component
        :type component: int or str
        :return:
        """

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        final_url = self.rules_by_templates_url.format(org_id, pipeline_id, component)

        result = self.session.get(final_url, headers=header)

        if result.status_code == 200:
            return json.loads(result.content)

        raise Exception(repr(result.content))

    @single_org
    def get_distinct_values_map_attributes(self, org_id=None, *, pipeline_id, flows, columns):
        """
        Retrieves distinct values from the given flow id

        :param org_id: Org id where the pipeline/flow is located
        :type org_id: int or str
        :param pipeline_id: Id of the pipeline to be queried
        :type pipeline_id: int or str
        :param flows: Flow id
        :type flows: int or str
        :param columns: Columns to used
        :type columns: list of str
        :return:
        """

        columns.append("TEMPLATE *")

        columns.append("SENSOR *")

        cols = []

        for col_name in columns:
            cols.append(('col', col_name))

        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        final_url = self.distinct_values_map_attributes.format(org_id, pipeline_id, flows)

        result = self.session.get(final_url, headers=header, params=cols)

        if result.status_code == 200:
            return json.loads(result.content)

        raise Exception(repr(result.content))

    @single_org
    def get_pipelines_v2(self, org_id=None, deleted=False, pipeline_type="standard"):
        """
        Retrieves the pipeline list from the given org

        :param org_id: Org id to be queried
        :type org_id: int or str
        :param deleted: Deletion status fo pipeliens to be queried
        :type deleted: bool or None
        :param pipeline_type: Pipeline type standard or function
        :type pipeline_type: str
        :return:
        """

        if deleted is True:
            query = self.gql_builder.get_deleted_pipelines(pipeline_type=pipeline_type)
        elif deleted is False:
            query = self.gql_builder.get_non_deleted_pipelines(pipeline_type=pipeline_type)
        else:
            query = self.gql_builder.get_pipelines_query(pipeline_type=pipeline_type)

        get_pipelines_request = self.graph_ql_query(org_id=org_id, query=query)

        if get_pipelines_request.status_code in range(200, 202):
            data = json.loads(get_pipelines_request.content)
            if "data" in data:
                if "artifacts" in data["data"]:
                    return data["data"]["artifacts"]

            return []
        raise Exception(repr(get_pipelines_request.content))

    @single_org
    def get_pipelines_functions(self, org_id=None, deleted=False):
        """
        Retrieves the pipeline list from the given org

        :param org_id: Org id to be queried
        :type org_id: int or str
        :param deleted: Deletion status fo pipeliens to be queried
        :type deleted: bool or None
        :return:
        """
        return self.get_pipelines_v2(org_id=org_id, deleted=deleted, pipeline_type="function")

    @single_org
    def verify_if_pipeline_exists_and_get_id(self, org_id=None, *, pipeline_name):
        """
        Verifies that the pipeline name exists on the given org. Returns pipeline id.

        :param org_id: Org id of pipeline
        :type org_id: int or str
        :param pipeline_name: Name of pipeline
        :type pipeline_name: str
        :return:
        """

        if pipeline_name is None:
            raise Exception("Pipeline name cant be none")

        pipeline_list = self.get_pipelines_v2(org_id=org_id)

        results = {"pipeline_id": None}

        for pipe in pipeline_list:
            if pipe["name"] == pipeline_name:
                existing_pipeline_id = pipe["id"]["id"]
                results.update({"pipeline_id": existing_pipeline_id})
                break

        return results

    @single_org
    def pipeline_exists(self, org_id=None, *, pipeline_name):
        """
        Verifies that the pipeline name exists on the given org

        :param org_id: Org to query if the pipeline exists
        :type org_id: int or str
        :param pipeline_name: Pipeline name to be queried
        :type pipeline_name: str
        :return:
        """
        if pipeline_name is None:
            raise Exception("Pipeline name cant be none")

        return self.verify_if_pipeline_exists_and_get_id(
            org_id=org_id,
            pipeline_name=pipeline_name
        )["pipeline_id"] is not None

    @single_org
    def update_pipeline_metadata(self, org_id=None, *, pipeline_id=None, name=None, description=None, facets=None):
        """
        Method to update a pipeline's metadata

        :param org_id: org id where the pipeline is stored
        :param pipeline_id: id of the pipeline to be edited
        :param name: new pipeline name
        :param description: description to be added
        :param facets:
        :return:
        """

        return self.update_artifact_metadata(
            org_id=org_id,
            artifact_id=pipeline_id,
            artifact_type="pipeline",
            name=name,
            description=description,
            facets=facets
        )

    @single_org
    def preview_flow(self, org_id=None, *, pipeline_id, flow_id):

        """
        Previews the given flow with default ordering
        :param org_id:
        :param pipeline_id:
        :param flow_id:
        :return:
        """

        header = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        cookies = {'_gid': 'GA1.1.1649742346.1580648966'}

        schema_url = self.flow_preview_url.format(org_id, pipeline_id, flow_id)

        flow_data = self.session.get(
            schema_url,
            headers=header,
            cookies=cookies
        )

        if flow_data.status_code in [200, 201, 202]:
            return json.loads(flow_data.content)

        raise Exception(repr(flow_data.content))

    @single_org
    def preview_flow_by_column(self, org_id=None, *, pipeline_id, flow_id, sort_column):

        """
        Previews the given flow ordered by the column name give
        :param org_id:
        :param pipeline_id:
        :param flow_id:
        :param sort_column:
        :return:
        """

        header = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        schema_url = self.preview_data_pipeline.format(org_id, pipeline_id, flow_id, sort_column)

        request = self.session.get(schema_url, headers=header)

        if request.status_code in [200, 201, 202]:
            return json.loads(request.content)

        raise Exception(repr(request.content))

    @single_org
    def flow_row_count(self, org_id=None, *, pipeline_id=None, flow_id=None):

        """
        Retrieves the number of rows the given flow contains
        :param org_id:
        :param pipeline_id:
        :param flow_id:
        :return:
        """

        header = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        schema_url = self.count_endpoint.format(org_id, pipeline_id, flow_id)

        request = self.session.get(schema_url, headers=header)

        if request.status_code in [200, 201, 202]:
            return json.loads(request.content)

        raise Exception(repr(request.content))

    @single_org
    def publish_pipeline(self, org_id=None, *, pipeline_id):

        """
        Published pipeline
        :param org_id:
        :param pipeline_id:
        :return:
        """

        header = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        run_url = self.run_pipeline_url.format(org_id, pipeline_id)

        params = [("parquet", True)]

        run_request = self.session.post(run_url, headers=header, params=params)

        if run_request.status_code in [200, 201, 202]:
            return json.loads(run_request.content)

        raise Exception(repr(run_request.content))

    @single_org
    def get_flow_distinct_values(self, org_id=None, *, pipeline_id, flow_id, col, sort_key, count=False):
        """
        Returns distinct values from a given flow, if count is true, it will return the count per distinct value
        :param org_id:
        :param pipeline_id:
        :param flow_id:
        :param col:
        :param sort_key:
        :param count:
        :return:
        """
        headers = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        if count:
            get_distinct_vals = self.distinct_count_values_url.format(org_id, pipeline_id, flow_id)
        else:
            get_distinct_vals = self.distinct_values_url.format(org_id, pipeline_id, flow_id)

        data = [("col", col), ("sortKey", sort_key), ("sortDir", "ASC")]

        result = self.session.get(get_distinct_vals, headers=headers, params=data)

        if result.status_code in [200, 201, 202]:
            return json.loads(result.content)

        raise Exception(repr(result.content))

    @single_org
    def autosync_pipeline(self, org_id=None, *, pipeline_id, autoSyncEnabled=True):

        """
        Changes the autosync status of a pipeline
        :param org_id:
        :param pipeline_id:
        :param autoSyncEnabled:
        :return:
        """
        headers = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        payload = {"autoSyncEnabled": autoSyncEnabled}

        post_pipeline_request = self.session.put(
            self.autosync_url.format(org_id, pipeline_id),
            headers=headers,
            data=json.dumps(payload)
        )

        if post_pipeline_request.status_code in [200, 201, 202]:
            return json.loads(post_pipeline_request.content)

        raise Exception(repr(post_pipeline_request.content))

    @single_org
    def get_pipeline_run_status(self, org_id=None, *, pipeline_id):
        """
        Returns distinct values from a given flow, if count is true, it will return the count per distinct value
        :param org_id:
        :param pipeline_id:
        :param flow_id:
        :param col:
        :param sort_key:
        :param count:
        :return:
        """

        header = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        status_url = self.run_pipeline_status_url.format(org_id, pipeline_id)

        run_request = self.session.get(status_url, headers=header)

        return run_request.status_code, json.loads(run_request.content)

    @single_org
    def wait_for_pipeline(self, org_id=None, *, pipeline_id):
        """
        Wait for a pipeline to be ready to be used

        :param org_id: Organization id where the pipeline was created
        :param pipeline_id: pipeline id number
        :return:
        """

        return self.wait_for_artifact(
            org_id=org_id,
            artifact_id=pipeline_id,
            artifact_type="pipeline"
        )

    @single_org
    def group_components(self, org_id=None, *, pipeline_id, components: list = None):
        """
        Group pipeline components

        :param org_id: org id where the pipeline is stored
        :param pipeline_id: id of the pipeline to be edited
        :param components: list of pipeline components ids to be grouped
        :return:
        """

        if components is None or len(components) == 0:
            raise Exception("Component list must have at least one ID")

        headers = self.build_header(
            org_id=str(org_id),
            aut_token=self.props.get_auth_token(cluster=self.cluster),
            others=self.content_type_header
        )

        payload = {"componentIds": components}

        post_pipeline_request = self.session.post(
            self.group_components_url.format(org_id, pipeline_id),
            headers=headers,
            data=json.dumps(payload)
        )

        if post_pipeline_request.status_code in self.OK:
            return json.loads(post_pipeline_request.content)

        raise Exception(repr(post_pipeline_request.content))



import json


class FlowColumn:

    def __init__(self, name=None, dataType=None, isOpt=None, columnType=None):
        self.name: str = name
        self.dataType: str = dataType
        self.isOpt: bool = isOpt
        self.columnType: str = columnType

    def build(self):
        return {
            "name": self.name,
            "dataType": self.dataType,
            "isOpt": self.isOpt,
            "columnType": self.columnType
        }



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unify.permissions import Permissions
from unify.helpers.Permissions import ArtifactType
from unify.helpers.Permissions import Verbs, Domains, Effects, DisplayNames
import json
import os
import uuid
from tests import test_org, cluster_name, props, test_dataset


class TestTemplates(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.permissions = Permissions(
            cluster=cluster_name,
            props=props
        )

    def test_get_org_config(self):

        try:
            response = self.permissions.get_org_config(org_id=test_org)

            assert True, ""
        except Exception as e:
            assert False, str(e)

    def test_get_rules(self):
        try:
            response = self.permissions.get_rules(org_id=test_org)
            assert True, ""
        except Exception as e:
            assert False, str(e)

    def test_get_user_rules(self):
        try:
            response = self.permissions.get_user_rules(org_id=test_org, user_id=1)

            assert True, ""
        except Exception as e:
            assert False, str(e)

    def test_get_dataset_rules(self):

        try:
            self.permissions.get_artifact_rules(
                org_id=test_org,
                artifact_type=ArtifactType.dataset,
                artifact_id=test_dataset["id"]
            )

            assert True, "Get datasets rule"

        except Exception as e:
            assert False, str(e)

    def test_selector_definitions(self):
        try:
            response = self.permissions.get_selector_definitions(org_id=test_org)
            assert True, ""
        except Exception as e:
            assert False, str(e)

    def test_check_permission(self):
        try:
            response = self.permissions.check_permission(
                org_id=test_org,
                artifact_type=ArtifactType.dataset,
                artifact_id=test_dataset["id"],
                verb=Verbs.write,
                user_id=1
            )
            assert True, ""
        except Exception as e:
            assert False, str(e)

    def test_get_selector_definition(self):
        try:
            response = self.permissions.get_selector_definition(
                org_id=test_org,
                selector_id=1
            )

            assert True, ""
        except Exception as e:
            assert False, str(e)

    def test_add_rule(self):
        try:
            response = self.permissions.add_rule(
                org_id=test_org,
                domain=Domains.dataset,
                effect=Effects.allow,
                verb=Verbs.write,
                userSelector=1,
                resourceSelector=3
            )

            assert True, ""
        except Exception as e:
            assert False, str(e)

    def test_get_user_specific_connector(self):
        response = self.permissions.get_specific_selectors(org_id=test_org, user=True)

        assert len(response) > 0, "No selectors found for users"

    def test_get_resource_specific_connector(self):
        response = self.permissions.get_specific_selectors(org_id=test_org, user=False)

        assert len(response) > 0, "No selectors found for resources"

    def test_get_user_all_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=True,
            display_name=DisplayNames.everyone
        )

        assert len(response) > 0, "No selectors found for all users"

    def test_get_user_none_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=True,
            display_name=DisplayNames.nobody
        )

        assert len(response) > 0, "No selectors found for no userss"

    def test_get_all_resources_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=False,
            display_name=DisplayNames.everything
        )

        assert len(response) > 0, "No selectors found for all resources"

    def test_get_none_resources_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=False,
            display_name=DisplayNames.nothing
        )
        assert len(response) > 0, "No selectors found for all resources"

    def test_negative_get_user_all_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=True,
            display_name=DisplayNames.everyone
        )

        assert response == self.permissions.get_everyone_user_selectors(
            org_id=test_org), "No selectors found for all users"

    def test_negative_get_user_none_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=True,
            display_name=DisplayNames.nobody
        )

        assert response == self.permissions.get_nobody_user_selectors(
            org_id=test_org), "No selectors found for all users"

    def test_negative_get_all_resources_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=False,
            display_name=DisplayNames.everything
        )

        assert response == self.permissions.get_everything_resource_selectors(
            org_id=test_org), "No selectors found for all users"

    def test_negative_get_none_resources_specific_connector(self):
        response = self.permissions.get_specific_selectors(
            org_id=test_org,
            user=False,
            display_name=DisplayNames.nothing
        )

        assert response == self.permissions.get_nothing_resource_selectors(
            org_id=test_org), "No selectors found for all users"

    def test_delete_rule(self):
        response = self.permissions.add_rule(
            org_id=test_org,
            domain=Domains.dataset,
            effect=Effects.allow,
            verb=Verbs.write,
            userSelector=1,
            resourceSelector=3
        )

        resp_delete = self.permissions.delete_rule(
            org_id=test_org,
            rule_id=response["data"]["permissions"]["addRule"]
        )

        assert resp_delete["data"]["permissions"]["deleteRule"], resp_delete



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains methods to interact with Unify access
"""

import time
from urllib.parse import urlparse
import pkg_resources
from unify.apirequestsmng import ApiRequestManager
from unify.helpers import staticsql


class AssetAccess(ApiRequestManager):
    """
    Class to interact with Unify access api client
    """

    def __init__(self, cluster, orgid, props):
        """
        Unify access api client class constructor

        :param cluster: Cluster name
        :type cluster: str
        :param orgid: Organization id
        :type orgid: int or str
        """
        super().__init__(cluster=cluster, props=props, org_id=orgid)

        required = {'psycopg2-binary'}
        installed = {pkg.key for pkg in pkg_resources.working_set}
        missing = required - installed

        if 'psycopg2-binary' in missing:
            raise Exception(
                "Please install {} by running pip install psycopg2-binary".format(
                    missing
                )
            )

        self.epoch_time = int(time.time())

        self.pi_tag_export_limit = {"piTagExportLimit": 999}

        self.expiry = {"expiry": 999}

        self.org_id = orgid

        result = urlparse(self.props.get_remote(cluster=self.cluster))

        self.connect_string = "host='{}' dbname='asset_hub_{}' port=8888 sslmode='require' user='{}' password='{}'".format(
            result.netloc,
            orgid,
            self.props.get_username(cluster),
            self.props.get_password(cluster)
        )

    def execute_query(self, query):
        """
        Creates a connection to Unify Access and executes the given query

        :param query: SQL query
        :type query: str
        :return:
        """
        required = {'psycopg2-binary'}
        installed = {pkg.key for pkg in pkg_resources.working_set}
        missing = required - installed

        if 'psycopg2-binary' in missing:
            raise Exception(
                "Please install {} by running pip install psycopg2-binary".format(
                    missing
                )
            )

        import psycopg2 as pg

        connection = pg.connect(self.connect_string)
        cursor = connection.cursor()

        cursor.execute(query=query)
        records = cursor.fetchall()
        output = []
        colnames = [desc[0] for desc in cursor.description]
        for record in records:
            aux = {}
            index = 0
            for desc in colnames:
                aux[desc] = record[index]
                index += 1

            output.append(aux)

        connection.close()

        return output

    def get_all_tables(self):
        """
        Retrieves all the tables that exists in org

        :return:
        """
        oid = self.execute_query(query=staticsql.OID_QUERY.format(self.org_id))

        if len(oid) < 1:
            raise Exception("No usable databases/datasets found for this org")

        return self.execute_query(query=staticsql.ALL_TABLES.format(oid[0]['oid']))



import json

from unify.helpers.graph_ql.graphql import GraphQLBuilder


class DatasetGrapql(GraphQLBuilder):

    def build_dataset_query(
            self,
            include_schema=True,
            count=True,
            page_num=1,
            page_size=100,
            is_deleted=False,
            facets=None,
            other_filters: dict = None
    ):

        art_count = self.build_generic_query(
            query_name="artifactCount",
            query_input={"pagination": "$pagination"},
        )

        artifacts = self.build_artifacts_query(pipeline=False)

        common_parts = self.build_common_parts_fragment()

        LastUpdatedParts = self.build_last_updated_fragment()

        fields = [
            "table",
            "numPipelines",
            "latestSequenceNumber",
            self.build_simple_query(
                query_name="labels",
                fields=[
                    'raw(fields: ["ean_source_type", "ean_ready", "ean_dataArchiveServerName"])',
                    "__typename"
                ]
            ),
            "__typename"
        ]

        if include_schema:
            fields.append("schema")

        DatasetParts = self.build_fragment(
            fragment_name="DatasetParts",
            interface_name="Dataset",
            fields=fields
        )

        queries = [artifacts]

        if count:
            queries.append(art_count)

        final = self.build_generic_query(
            operation_name="fetchDatasets",
            operation_type="query",
            operation_input={"$pagination": "PaginationInput!"},
            operation_queries=queries
        )

        final_query = "".join(
            [common_parts, LastUpdatedParts, DatasetParts, final]
        )

        filters = 'isDeleted = "{}" && type = "dataset"'.format(is_deleted)

        if facets:
            for facet in facets:
                filters += ' && facets = "{}"'.format(str(facet))

        if other_filters:
            for key_filter, value_filter in other_filters.items():
                filters += ' && ({} ~ \"{}\")'.format(key_filter, value_filter)

        results = {
            "operationName": "fetchDatasets",
            "query": final_query,
            "variables": {
                "pagination": {
                    "pageInfo": {
                        "pageNum": page_num,
                        "pageSize": page_size
                    },
                    "filter": filters
                }
            }
        }

        return json.dumps(results)



import json
from unify.helpers.datasets.column import DatasetHeader
from unify.helpers.pipeline.flow_column import FlowColumn


class FlowSchema:

    def __init__(self, columns: list = None):
        self.columns: list = []
        if columns is not None:
            self.columns = columns.copy()

    def to_dataset_schema(self):
        dataset_schema = []

        for col in self.columns:
            flow_col = FlowColumn(**col)
            dataset_schema.append(DatasetHeader.from_pipline_flow_column(flow_col))

        return dataset_schema



import json
from unify.helpers.pipeline.flow_column import FlowColumn

"""
  {
    "header": "name",
    "column": {
      "type": "text",
      "properties": {
        "optional": false,
        "datasets.v1.columnType": "Normal"
      }
    }
  }
"""

col_type_translation = {
    "id": "Normal",
    "normal": "Normal",
    "Normal": "Normal",
    "template": "Template",
    "sensor": "Sensor",
    "equipmentid": "EquipmentId",
    "empty": "Empty",
    "attributeid": "AttributeId",
    "primaryid": "PrimaryId",
    "unitofmeasure": "UnitOfMeasure",
    "sourceserver": "SourceServer"
}

# https://github.com/Voltir/datum/blob/master/core/src/main/scala/datum/patterns/schemas/Type.scala#L18
col_data_translation = {
    "STRING": "text",
    "BOOLEAN": "boolean",
    "LONG": "long",
    "DOUBLE": "double",
    "INT": "int",
    "FLOAT": "float"
}


class DatasetColumn:

    def __init__(self, data_type: str = None, optional: bool = False, column_type: str = None):
        self.data_type: str = data_type
        self.optional: bool = optional
        self.column_type: str = column_type

    def build(self):
        return {
            "type": self.data_type,
            "properties": {
                "optional": self.optional,
                "datasets.v1.columnType": self.column_type
            }
        }


class DatasetHeader:

    @staticmethod
    def from_pipline_flow_column(flow_column: FlowColumn):
        dataset_column = DatasetColumn(
            data_type=col_data_translation[flow_column.dataType],
            optional=flow_column.isOpt,
            column_type=col_type_translation[flow_column.columnType.lower()]
        )

        return DatasetHeader(
            name=flow_column.name,
            column=dataset_column
        ).build()

    def __init__(self, name: str = None, column: DatasetColumn = None):
        self.header: str = name
        self.column = column if column is not None else DatasetColumn()

    def build(self):
        return {
            "header": self.header,
            "column": self.column.build()
        }



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from unify.templates import Templates
import json
import os
import uuid
from tests import test_org, cluster_name, props


class TestTemplatesWithDefaultOrg(unittest.TestCase):
    templates = Templates(cluster_name, props, org_id=test_org)
    templates.upload_string_content_file(
        test_org,
        """Template Name,Attribute,Datatype,UoM,Attribute Type
        Conveyer,Vibration,Float4,hz,""")
    test_template = templates.list_asset_templates()[0]
    template_id = test_template['id']
    template_name = test_template['name']
    version = test_template['version']

    def test_category(self):
        categories = ['python sdk test', 'antoher category']
        self.templates.category(template_id=self.template_id, template_name=self.template_name, version=self.version,
                                categories=categories)

        all_categories = self.templates.list_all_categories()
        self.assertListEqual(list(all_categories.keys()), categories)

        template_category_ids = self.templates.get_template(template_id=self.template_id)['categoryIds']
        self.assertListEqual(list(all_categories.values()), template_category_ids)

    def test_category_with_existing_category(self):
        categories = ['python sdk test']
        self.templates.category(
            template_id=self.template_id,
            template_name=self.template_name,
            version=self.version,
            categories=categories
        )

        new_categories = ['python sdk test', 'antoher category']
        self.templates.category(
            template_id=self.template_id,
            template_name=self.template_name,
            version=self.version,
            categories=new_categories
        )

        all_categories = self.templates.list_all_categories()
        self.assertListEqual(list(all_categories.keys()), new_categories)

        template_category_ids = self.templates.get_template(template_id=self.template_id)['categoryIds']
        self.assertListEqual(list(all_categories.values()), template_category_ids)

    def test_add_attributes(self):
        name = str(uuid.uuid4())
        create = self.templates.create_template_attribute_params(
            template_id=self.template_id,
            attribute_name=str(uuid.uuid4()),
            sanitized_name=name.replace("-", "_"),
            data_type="Int32",
            attribute_type="limitLow",
            description=str(uuid.uuid4()),
            uom=str(uuid.uuid4()),
            interpolation="Lookback"
        )
        self.assertTrue("id" in create, create)

    def test_unit_of_measure_list(self):
        """
        Verify org has UOM list

        :return:
        """

        uom_list = self.templates.get_uom()

        self.assertTrue(len(uom_list["items"]) > 0)



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest
from tests import *


class TestPipelinesOverrideOrgId(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.pipeline = Pipelines(
            cluster=cluster_name,
            props=props
        )

    def test_create_pipeline(self):
        new_pipe = self.pipeline.create_pipeline(
            name=str(uuid.uuid4()),
            function=False,
            org_id=test_org
        )

        self.assertTrue('pipeline' in new_pipe, new_pipe)
        self.assertTrue('id' in new_pipe["pipeline"], new_pipe)
        self.assertTrue('data' in new_pipe["pipeline"], new_pipe)
        self.assertTrue('pipelineType' in new_pipe["pipeline"]["data"], new_pipe)
        self.assertEqual("standard", new_pipe["pipeline"]["data"]["pipelineType"], new_pipe)

    def test_create_function(self):
        new_pipe = self.pipeline.create_pipeline(
            name=str(uuid.uuid4()),
            function=True,
            org_id=test_org
        )

        self.assertTrue('pipeline' in new_pipe, new_pipe)
        self.assertTrue('id' in new_pipe["pipeline"], new_pipe)
        self.assertTrue('data' in new_pipe["pipeline"], new_pipe)
        self.assertTrue('pipelineType' in new_pipe["pipeline"]["data"], new_pipe)
        self.assertEqual("function", new_pipe["pipeline"]["data"]["pipelineType"], new_pipe)

    def test_get_pipelines_v2(self):
        pipe_list = self.pipeline.get_pipelines_v2(org_id=test_org)

        self.assertTrue(len(pipe_list) > 0, pipe_list)

    def test_get_pipeline(self):

        try:
            pipe_list = self.pipeline.get_pipeline(
                pipeline_id=test_pipeline["pipeline"]["id"],
                org_id=test_org
            )
            self.assertTrue(True, pipe_list)
        except Exception as e:
            self.assertTrue(False, e)

    def test_regular_duplicate(self):
        duplicate = self.pipeline.regular_duplicate(
            pipeline_id=test_pipeline["pipeline"]["id"],
            new_name=str(uuid.uuid4()),
            org_id=test_org
        )
        self.assertTrue("id" in duplicate, duplicate)

    def test_regular_duplicate_id_required(self):

        try:
            _ = self.pipeline.regular_duplicate(
                pipeline_id=test_pipeline["pipeline"]["id"],
                org_id=test_org
            )
            self.assertTrue(False, "Duplicate must require pipeline id")
        except Exception as e:

            self.assertTrue(True, e)

    def test_regular_duplicate_name_required(self):

        try:
            _ = self.pipeline.regular_duplicate(
                pipeline_id=test_pipeline["pipeline"]["id"],
                org_id=test_org
            )
            self.assertTrue(False, "Duplicate must require new name")
        except Exception as e:

            self.assertTrue(True, e)

    def test_verify_if_pipeline_exists_and_get_id(self):
        pipeline_id = self.pipeline.verify_if_pipeline_exists_and_get_id(
            pipeline_name=test_pipeline["pipeline"]["data"]["name"],
            org_id=test_org
        )

        self.assertTrue("pipeline_id" in pipeline_id)
        self.assertTrue(pipeline_id["pipeline_id"] is not None)

    def test_verify_if_pipeline_exists_and_get_id_negative(self):
        pipeline_id = self.pipeline.verify_if_pipeline_exists_and_get_id(
            pipeline_name="{}{}".format(str(uuid.uuid4()), str(uuid.uuid4())),
            org_id=test_org
        )

        self.assertTrue("pipeline_id" in pipeline_id)
        self.assertTrue(pipeline_id["pipeline_id"] is None)

    def test_verify_if_pipeline_exists_and_get_id_name_required(self):
        try:
            _ = self.pipeline.verify_if_pipeline_exists_and_get_id(org_id=test_org)
            self.assertTrue(True, "verify_if_pipeline_exists_and_get_id method requires pipeline name")
        except Exception as e:
            self.assertTrue(True, e)

    def test_pipeline_exists(self):

        self.assertTrue(
            self.pipeline.pipeline_exists(
                pipeline_name=test_pipeline["pipeline"]["data"]["name"],
                org_id=test_org
            ),
            "Pipeline does not exists"
        )

    def test_pipeline_exists_negative(self):

        self.assertFalse(
            self.pipeline.pipeline_exists(
                pipeline_name="{}{}".format(str(uuid.uuid4()), str(uuid.uuid4())),
                org_id=test_org
            ),
            "Non existing pipeline exists"
        )

    def test_pipeline_exists_name_required(self):
        try:
            _ = self.pipeline.pipeline_exists(org_id=test_org)
            self.assertTrue(True, "pipeline_exists method requires pipeline name")
        except Exception as e:
            self.assertTrue(True, e)

    def test_rename_pipeline(self):
        new_pipe = self.pipeline.create_pipeline(
            name=str(uuid.uuid4()),
            function=False,
            org_id=test_org
        )
        self.assertTrue('id' in new_pipe["pipeline"], new_pipe)

        try:
            _ = self.pipeline.update_pipeline_metadata(
                org_id=test_org,
                pipeline_id=new_pipe["pipeline"]["id"],
                name="{}__renamed".format(str(uuid.uuid4()))
            )
            self.assertTrue(True, "Pipeline was renamed")
        except Exception as e:
            self.assertTrue(False, e)

    def test_add_facets_pipeline(self):
        original_name = "Add facets {}".format(str(uuid.uuid4())[:10])
        new_pipe = self.pipeline.create_pipeline(
            name=original_name,
            function=False,
            org_id=test_org
        )

        try:
            _ = self.pipeline.update_pipeline_metadata(
                org_id=test_org,
                pipeline_id=new_pipe["pipeline"]["id"],
                facets=[str(uuid.uuid4())[:10]]
            )
            self.assertTrue(True, "Facets were added to Pipeline")
        except Exception as e:
            self.assertTrue(False, e)

    def test_add_description_pipeline(self):
        original_name = "Add description {}".format(str(uuid.uuid4())[:10])
        new_pipe = self.pipeline.create_pipeline(
            name=original_name,
            function=False,
            org_id=test_org
        )

        try:
            _ = self.pipeline.update_pipeline_metadata(
                org_id=test_org,
                pipeline_id=new_pipe["pipeline"]["id"],
                description=str(uuid.uuid4())
            )

            self.assertTrue(True, "Description added to Pipeline")
        except Exception as e:
            self.assertTrue(False, e)

    def test_wait_for_pipeline(self):
        """
        Verify waitfor pipeline endpoint with default org
        :return:
        """
        try:
            _ = self.pipeline.wait_for_pipeline(
                pipeline_id=test_pipeline["pipeline"]["id"],
                org_id=test_org
            )

            self.assertTrue(True, "Pipeline became ready")
        except Exception as e:
            self.assertTrue(False, e)

    def test_group_components(self):
        original_name = "Group components {}".format(str(uuid.uuid4())[:10])
        new_pipe = self.pipeline.create_pipeline(
            name=original_name,
            function=False,
            org_id=test_org
        )

        self.pipeline.wait_for_pipeline(
            org_id=test_org,
            pipeline_id=new_pipe["pipeline"]["id"]
        )

        self.pipeline.update_pipeline_from_json(
            update_payload=json.loads(open("tests/data/pipeline.json", "r+").read()),
            pipeline_name=original_name,
            pipeline_id=new_pipe["pipeline"]["id"],
            org_id=test_org,
        )

        try:
            response = self.pipeline.group_components(
                org_id=test_org,
                pipeline_id=new_pipe["pipeline"]["id"],
                components=["44564", "674714"]
            )
            self.assertTrue(True, response)
        except Exception as e:
            self.assertTrue(False, e)



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains helper methods for files
"""
import io
import csv


def json_to_csv(data_array, io_format=False):
    """
    Converts a json array into a csv file

    :param data_array: JSON list
    :type data_array: list
    :param io_format: Flag to return IO format
    :type io_format: bool, optional
    :return:
    """
    output = io.StringIO()

    writer = csv.writer(output, quoting=csv.QUOTE_ALL, delimiter=',')

    first = True
    columns = []
    for value in data_array:
        if first:
            first = False
            writer.writerow(value.keys())
            columns = value.keys()

        new_row = []
        for col in columns:
            new_row.append(value[col])
        writer.writerow(new_row)

    if io_format:
        return output

    return output.getvalue()


def csv_to_json(csv_data, encoding="UTF-8"):
    """
    Creates a json object from a csv data

    :param csv_data: Contents in CSV format
    :type csv_data: str
    :param encoding: Encoding content
    :type csv_data: str
    :return:
    """
    csv_file = []
    reader = csv.DictReader(io.StringIO(csv_data.decode(encoding)))
    for row in reader:
        csv_file.append(row)

    return csv_file


def stream_iterable(container, chunk):
    """
    Generates sub chunks of the given size on the given iterable

    :param container: interable object
    :type container: interable object
    :param chunk: sub chunk size to be create
    :type chunk: int
    :return:
    """
    counter = 0
    ex = False
    while not ex:

        yield container[counter:counter + chunk]
        counter += chunk

        if counter > len(container):
            ex = True


def create_schema_dataset(csv_data, name):
    """
    Creates the schema to be used on on static route create dataset

    :param csv_data: CSV data from the source to be used
    :type csv_data: str
    :param name: Dataset name being used
    :type name: str
    :return:
    """
    schema = {
        "name": name,
        "cause": [],
        "schema": {
            "properties": {},
            "columns": []
        }
    }
    reader = csv.DictReader(io.StringIO(csv_data))
    headers = next(reader)
    for header in headers:
        schema["schema"]["columns"].append(
            {
                "column": {
                    "type": "text",
                    "properties": {}
                },
                "header": header
            }
        )

    return schema



import json
from unify.helpers.graph_ql.graphql import GraphQLBuilder


class PipelineGrapql(GraphQLBuilder):

    def get_non_deleted_pipelines(self, pipeline_type="standard"):
        """
        Creates graphql query to retrieve all non deleted pipelines

        :param pipeline_type: function or standard
        :return:
        """
        return self.build_pipeline_query(is_deleted=False, pipeline_type=pipeline_type)

    def get_deleted_pipelines(self, pipeline_type="standard"):
        """
        Creates graphql query to retrieve all deleted pipelines

        :param pipeline_type:
        :return:
        """
        return self.build_pipeline_query(is_deleted=True, pipeline_type=pipeline_type)

    def get_pipelines_query(self, pipeline_type="standard"):
        """
        Creates graphql query to retrieve all non deleted standard pipelines

        :param pipeline_type:
        :return:
        """
        return self.build_pipeline_query(is_deleted=None, pipeline_type=pipeline_type)

    def get_functions_query(self):
        """
        Creates graphql query to retrieve all non deleted function pipelines

        :return:
        """
        return self.build_pipeline_query(
            pipeline_type="function"
        )

    def build_pipeline_query(
            self,
            count=True,
            page_num: int = 1,
            page_size: int = 100,
            facets=None,
            is_deleted: object = False,
            pipeline_type="standard"
    ):

        if is_deleted not in [False, True, None]:
            raise Exception("is_deleted must be False,True,None")

        artifacts = self.build_artifacts_query(pipeline=True)

        common_parts = self.build_common_parts_fragment()

        last_updated_parts = self.build_last_updated_fragment()

        filter_relations = 'direction=forward && type=uses && artifact.type=dataset'

        if is_deleted is True:
            filter_relations += " && artifact.isDeleted = true"
        elif is_deleted is False:
            filter_relations += " && artifact.isDeleted = false"

        fields = [
            "legacyId",
            "autosync",
            "pipelineType",
            "columnFunction",
            self.build_simple_query(
                query_name="relationships",
                query_alias="sources",
                fields=[
                    self.build_simple_query(
                        query_name="artifact",
                        fields=[
                            "name",
                            "__typename"
                        ]
                    ),
                    "__typename"
                ],
                query_input={
                    "pagination": {
                        "filter": filter_relations,
                        "pageInfo": {
                            "pageNum": page_num,
                            "pageSize": page_size
                        }
                    }
                }
            )
            ,
            self.build_simple_query(
                query_name="relationships",
                query_alias="sinks",
                fields=[
                    self.build_simple_query(
                        query_name="artifact",
                        fields=[
                            "name",
                            "__typename"
                        ]
                    ),
                    "__typename"
                ],
                query_input={
                    "pagination": {
                        "filter": filter_relations,
                        "pageInfo": {
                            "pageNum": page_num,
                            "pageSize": page_size
                        }
                    }
                }
            ),
            "__typename"
        ]

        dataset_parts = self.build_fragment(
            fragment_name="PipelineParts",
            interface_name="Pipeline",
            fields=fields
        )

        queries = [artifacts]

        if count:
            art_count = self.build_generic_query(
                query_name="artifactCount",
                query_input={"pagination": "$pagination"},
            )
            queries.append(art_count)

        final = self.build_generic_query(
            operation_name="fetchPipelines",
            operation_type="query",
            operation_input={"$pagination": "PaginationInput!"},
            operation_queries=queries
        )

        final_query = ''.join(
            [final, common_parts, last_updated_parts, dataset_parts]
        )

        filters = 'type = "pipeline" && pipelineType="{}"'.format(pipeline_type)

        if is_deleted is True:
            filters += ' && isDeleted = "true"'
        elif is_deleted is False:
            filters += ' && isDeleted = "false"'

        if facets:
            for facet in facets:
                filters += ' && facets = "{}"'.format(str(facet))

        results = {
            "operationName": 'fetchPipelines',
            "query": final_query,
            "variables": {
                "pagination": {
                    "filter": filters,
                    "pageInfo": {
                        "pageNum": page_num,
                        "pageSize": page_size
                    }
                }
            }
        }

        return json.dumps(results).replace("'", "")



import json
from unify.helpers.graph_ql.graphql import GraphQLBuilder
from unify.helpers.Permissions import ArtifactType
from unify.helpers.Permissions import Verbs, Domains, Effects


class PermissionsGraphQl(GraphQLBuilder):

    def __init__(self):
        pass

    def build_get_org_config_query(self):
        return self.build_generic_query(
            query_name="permissions",
            operation_type="query",
            fields=[self.build_simple_query(
                query_name="getOrgConfig",
                fields=["orgId", "globalRuleOptOut"]
            )]
        )

    def get_org_config(self, ):
        return self.build_response(
            query=self.build_get_org_config_query()
        )

    def build_user_selector(self):
        return self.build_generic_query(
            query_name="userSelector", fields=["id", "selectorType", "org", "displayName", "selector"]
        )

    def build_resource_selector(self):
        return self.build_generic_query(
            query_name="resourceSelector",
            fields=["id", "selectorType", "org", "displayName", "selector"]
        )

    def build_get_rules_query(self):
        return self.build_generic_query(
            query_name="permissions",
            operation_type="query",
            fields=[self.build_simple_query(
                query_name="getRules",
                fields=[
                    "id", "domains", "scope", "verb", "effect",
                    self.build_user_selector(),
                    self.build_resource_selector()
                ]
            )]
        )

    def get_rules(self):
        return self.build_response(
            query=self.build_get_rules_query(),
        )

    def build_get_user_rules_query(self, user_id):
        queries = [
            self.build_simple_query(
                query_name="getUserRules",
                fields=[
                    "id", "scope", "effect", "verb", "domains",
                    self.build_user_selector(),
                    self.build_resource_selector()
                ],
                query_input={"userId": user_id}
            )
        ]

        return self.build_generic_query(
            query_name="permissions",
            operation_type="query",
            fields=queries,
        )

    def get_user_rules(self, user_id):
        return self.build_response(
            query=self.build_get_user_rules_query(user_id),
            variables={"userId": user_id}
        )

    def build_artifact_rules_query(self, artifact_type, artifact_id):
        queries = [
            self.build_simple_query(
                query_name="getArtifactRules",
                fields=[
                    "id", "scope", "effect", "verb", "domains",
                    self.build_user_selector(),
                    self.build_resource_selector()
                ],
                query_input={"artifactId": '"{}"'.format(str(artifact_id)), "at": artifact_type}
            )
        ]

        return self.build_generic_query(
            query_name="permissions",
            operation_type="query",
            fields=queries,
        )

    def get_artifact_rules(self, artifact_type, artifact_id):
        return self.build_response(
            query=self.build_artifact_rules_query(artifact_type, artifact_id),
        )

    def build_selector_definitions_query(self):
        return self.build_generic_query(
            query_name="permissions",
            operation_type="query",
            fields=[self.build_simple_query(
                query_name="getSelectorDefinitions",
                fields=[
                    "id", "org", "displayName", "selectorType", "selector"
                ]
            )]
        )

    def get_selector_definitions(self):
        return self.build_response(
            query=self.build_selector_definitions_query()
        )

    def build_check_query(self, artifact_type, artifact_id, user_id, verb):
        queries = [
            self.build_simple_query(
                query_name="check",
                fields=[
                    "allowed"
                ],
                query_input={
                    "artifactId": '"{}"'.format(str(artifact_id)),
                    "at": artifact_type,
                    "userId": user_id,
                    "verb": verb
                }
            )
        ]
        return self.build_generic_query(
            query_name="permissions",
            operation_type="query",
            fields=queries,
        )

    def check(self, artifact_type, artifact_id, user_id, verb):
        return self.build_response(
            query=self.build_check_query(artifact_type, artifact_id, user_id, verb)
        )

    def build_selector_definition_query(self, selector_id):
        queries = [
            self.build_simple_query(
                query_name="getSelectorDefinition",
                fields=[
                    "id", "org", "displayName", "selectorType", "selector"
                ],
                query_input={"id": selector_id}
            )
        ]

        return self.build_generic_query(
            query_name="permissions",
            operation_type="query",
            fields=queries,
        )

    def get_selector_definition(self, selector_id):
        return self.build_response(
            query=self.build_selector_definition_query(selector_id=selector_id)
        )

    def build_response(self, query, variables: dict = None, operation_name=None):
        obj = {
            "query": query
        }

        if operation_name:
            obj["operation_name"] = operation_name

        if variables:
            obj["variables"] = variables

        return json.dumps(obj)

    def build_rule_mutation(self, domain, verb, effect, userSelector, resourceSelector):
        sub_data = {
            "domains": domain,
            "verb": verb,
            "effect": effect,
            "userSelector": userSelector,
            "resourceSelector": resourceSelector
        }

        final = []

        for key, value in sub_data.items():
            final.append("{}:{}".format(key, value))

        data = {
            "rule": "{" + ' '.join(final) + "}"
        }

        simple = self.build_simple_query(
            query_name="addRule",
            query_input=data,
        )

        return self.build_response(
            query="mutation { permissions { " + simple + "}}"
        )

    def build_delete_rule_mutation(self, rule_id):

        simple = self.build_simple_query(
            query_name="deleteRule",
            query_input={"id": '"{}"'.format(rule_id)},
        )

        return self.build_response(
            query="mutation { permissions { " + simple + "}}"
        )


    def build_add_selector_mutation(self, domain, verb, effect, userSelector, resourceSelector):
        sub_data = {
            "domains": domain,
            "verb": verb,
            "effect": effect,
            "userSelector": userSelector,
            "resourceSelector": resourceSelector
        }

        final = []

        for key, value in sub_data.items():
            final.append("{}:{}".format(key, value))

        data = {
            "rule": "{" + ' '.join(final) + "}"
        }

        simple = self.build_simple_query(
            query_name="addRule",
            query_input=data,
        )

        return self.build_response(
            query="mutation { permissions { " + simple + "}}"
        )



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains general class to inherit from when creating a new api
client service
"""
import time

import json
import requests
from unify.helpers.graph_ql.graphql import GraphQLBuilder


class ApiRequestManager:
    """
    Class to interact with general api request methods
    """
    OK = [200, 201, 202]

    def __init__(self, cluster, props, org_id=None):
        """
        Class constructor

        :param cluster: Cluster name to interact with
        :type cluster: str
        :param props: Instantiated Properties from unify/properties
        :type props: class:`unify.properties.Properties`
        """
        self.props = props
        self.session = requests.Session()
        self.content_type_header = {"content-type": "application/json"}
        self.content_octet = {"content-type": "application/octet-stream"}
        self.upload_content_type_header = {"Content-Type": "multipart/form-data"}
        self.delete_content_type_header = {"Content-Type": "application/json"}
        self.different_content_type = {"'Content-Type'": "application/json"}

        self.list_templates_uri = 'api/assetTemplates'

        self.x_auth_token_header = "x-auth-token"
        self.x_org_header = "x-organization-id"
        self.cluster = cluster
        self.evergreen_enabled = True

        self.epoch_time = int(time.time())
        self.pi_tag_export_limit = {"piTagExportLimit": 999}
        self.expiry = {"expiry": None}

        self.org_id = org_id

        auth = self.props.get_auth_token(cluster=self.cluster)

        self.api_session_url = self.props.get_remote(cluster=self.cluster) + 'api/sessions'

        if auth is None or len(auth) == 0:
            self.props.set_auth_token(
                token=self.auth_token(),
                cluster=self.cluster
            )

        self.gql_url = self.props.get_remote(self.cluster) + "artifacts/graphql"

        self.upper_gql_builder = GraphQLBuilder()

    def close_session(self):
        """
        Closes request session

        :return:
        """
        self.session.close()

    def verify_auth_token(self, cluster_name):
        """
        Verifies if the auth token is set

        :return:
        """
        if self.props.get_auth_token(cluster=cluster_name) is None:
            raise Exception(
                "Authentication is required with remote server, login running ah login"
            )

    def verify_remote(self, cluster_name):
        """
        Verifies if the remote cluster is set

        :return:
        """
        if self.props.get_remote(cluster=cluster_name) is None:
            raise Exception(
                "Remote server must be setup first"
            )

    def verify_properties(self, cluster_name):
        """
        Verify both auth token and remote cluster

        :return:
        """
        self.verify_auth_token(cluster_name=cluster_name)
        self.verify_remote(cluster_name=cluster_name)

    def auth_token(self, query_params={}):
        """
        Authenticates the given credentials and retrieves auth token.

        Note: this method doesn't pass an x-auth-token header - it's a
        credential exchange to _get_ an auth-token.

        :return:
        """

        username = self.props.get_username(cluster=self.cluster)
        password = self.props.get_password(cluster=self.cluster)

        try:
            data = {"email": username, "password": password}

            post_response = self.session.post(
                self.api_session_url,
                headers=self.content_type_header,
                data=json.dumps(data),
                params=query_params
            )

            if post_response.status_code == 200:
                return json.loads(post_response.content)["authToken"]

            raise Exception(post_response.content)

        except Exception as error:

            raise error

    def graph_ql_query(self, org_id, query):
        """
        Generates a request to graphql query

        :param org_id: Organization id to be queried
        :param query: GraphQL Query body
        :return:
        """
        header = self.build_header(
            org_id=org_id,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )

        query_results = self.session.post(
            self.gql_url,
            headers=header,
            data=query
        )

        return query_results

    def wait_for_artifact(self, org_id, artifact_id, artifact_type):
        """
        Tells the servewr to wait for an artifact to be ready, request returns when artifact is ready to use

        :param org_id: Organization id to be queried
        :param artifact_id: Artifact Identification UUID
        :param artifact_type: Dataset or Pipeline
        :return:
        """

        query = self.upper_gql_builder.build_wait_for_query(artifact_id=artifact_id, artifact_type=artifact_type)

        get_pipelines_request = self.graph_ql_query(org_id=org_id, query=query)

        if get_pipelines_request.status_code in self.OK:
            data = json.loads(get_pipelines_request.content)
            if "data" in data:
                if "waitFor" in data["data"]:
                    return data["data"]["waitFor"]

            return []

        raise Exception(repr(get_pipelines_request.content))

    def update_artifact_metadata(self, org_id=None, *,artifact_type, artifact_id=None, name=None, description=None, facets=None):
        """
        Request to update metadata of an artifact

        :param org_id: Organization id where the artifact is stored
        :param artifact_type: Dataset or Pipeline
        :param artifact_id: Artifact Identification UUID
        :param name: New name to update
        :param description: New Description to update
        :param facets: New list of facets to be included
        :return:
        """

        query = self.upper_gql_builder.mutation_query(
            artifact_id=str(artifact_id),
            artifact_type=artifact_type,
            new_name=name,
            description=description,
            facets=facets
        )

        response = self.graph_ql_query(org_id=org_id, query=query)

        if response.status_code in self.OK:
            return json.loads(response.content)

        raise Exception(repr(response.content))

    def build_header(self, aut_token=None, org_id=None, others=None):
        """
        Creates request header

        :param aut_token: x-auth-token to be included
        :type aut_token: str
        :param org_id: Org id where the request will be aimed
        :type org_id: int or str
        :param others: Other headers to be included.
        :type others: dict
        :return:
        """
        header = {}

        if others is not None and isinstance(others, dict):
            header.update(others)

        if aut_token is not None:
            header[self.x_auth_token_header] = aut_token

        if org_id is not None:
            header[self.x_org_header] = str(org_id)

        return header

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def simple_get(self, final_url):

        header = self.build_header(
            others=self.content_type_header,
            aut_token=self.props.get_auth_token(cluster=self.cluster)
        )

        query_results = self.session.get(
            final_url,
            headers=header
        )

        return query_results



# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Contains methods to manage the properties that the SDK needs.
For example, credentials, cluster metadata.
"""

import json
import uuid
import threading
from urllib.parse import urlparse
from enum import Enum
from enum import unique
import keyring
from unify.clusters import Cluster


@unique
class ClusterSetting(Enum):
    """
    Class to store cluster data in memory
    """
    MEMORY = 1
    KEY_RING = 2
    FILE = 3

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class Properties:
    """
    Class respresenting cluster properties
    """

    _instance = None
    _lock = threading.Lock()

    def __new__(cls, clusterSetting=ClusterSetting.KEY_RING):
        """
        Override __new__ to implement singleton semantics. Note that this is also threadsafe.
        """
        if cls._instance is None:
            with cls._lock:
                if not cls._instance:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self, clusterSetting=ClusterSetting.KEY_RING):
        """
        Properties constructor

        :param clusterSetting: Cluster settings to decide if use keyring, memory or file
        :type clusterSetting: `unify.properties.Properties.Enum`
        """
        self._SYSTEM = "ah"
        self._REMOTE = "d0adae9c-a1b7-11ea-906d-acde48001122_{}"
        self._AUTHTOKEN = "c8f9465c-a1b7-11ea-906d-acde48001122_{}"
        self._USERNAME = "c2c731a4-a1b7-11ea-906d-acde48001122_{}"
        self._PASSWORD = "b901bcca-a1b7-11ea-906d-acde48001122_{}"
        self._NAMES = '23556a06-a1c0-11ea-906d-acde48001122'
        self._DEFAULT = '69a9e665-bdd0-40aa-b729-735c8997d3f6'
        self._ASSETSYNC = 'cc61e5c1-ab56-43cb-b848-f342eaafc90c_{}'

        self.clusters = {}
        self.clusterSetting = clusterSetting

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def set_default(self, name):
        """
        Sets default cluster

        :param name: Cluster name to set as default
        :type name: str
        """
        try:
            clus = keyring.get_password(self._SYSTEM, self._REMOTE.format(name))

            if clus is not None:
                keyring.set_password(self._SYSTEM, self._DEFAULT, name)
            else:
                msg = (
                    "Cluster {}  cant be found and was"
                    "not possible to set it as default".format(name)
                )
                raise Exception(msg)

        except Exception as err:
            raise err

    def get_default(self):
        """
        Returns the default cluster

        :return:
        """
        return keyring.get_password(self._SYSTEM, self._DEFAULT)

    def store_cluster(self, username, password, cluster, name=None, assetsync=True):
        """
        Stores cluster data. This method does not makes a login request.

        :param username: Username to be used as part of login credentials
        :type username: str
        :param password: Password to be used as part of login credentials
        :type password: str
        :param cluster: Cluster URL
        :type cluster: str
        :param name: Cluster name, unique identifier
        :type name: str, optional
        :param assetsync: Deprecated - Always set to true
        :type assetsync: bool, optional
        :return:
        """

        # Override assetsync to True - deprecated argument
        assetsync = True

        if name is None:
            name = str(uuid.uuid1())[0:10]

        if self.clusterSetting is ClusterSetting.MEMORY:

            self.clusters[name] = Cluster(
                userName=username,
                password=password,
                cluster=cluster,
                token=''
            )

        else:
            self.set_password(password=password, cluster=name)
            self.set_username(username=username, cluster=name)
            self.set_remote(server_uri=cluster, name=name)

    def get_memory_cluster(self, name):
        """
        Retrieves the cluster data from memory

        :param name: Cluster name
        :type name: str
        :return: hostname
        """
        cluster = self.clusters.get(name)

        if cluster is None:
            raise Exception('Remote server must be setup first by running')

        return cluster

    def set_asset_sync(self, assetsync, cluster):
        """
        Deprecated - this method does nothing, just returns.

        :param assetsync: New AssetSync Value
        :type assetsync: bool
        :param cluster: Cluster name
        :type cluster: str
        :return:
        """
        pass

    def get_asset_sync(self, cluster):
        """
        Deprecated - Returns the assetSync value of a given cluster, always True

        :param cluster: Cluster name/id
        :type cluster: str
        :return:
        """
        return True

    def set_remote(self, server_uri, name=None):
        """
        Overrides cluster URL

        :param server_uri: New cluster URL
        :type server_uri: str
        :param name: Cluster name/id
        :type name: str
        :return:
        """

        try:
            result = urlparse(server_uri)
            if result.scheme == "" or result.netloc == "":
                raise Exception("Invalid url used as Remote, please use a valid URL")

            if server_uri[-1] != "/":
                server_uri = "{}/".format(server_uri)

            if self.clusterSetting is ClusterSetting.KEY_RING:

                keyring.set_password(self._SYSTEM, self._REMOTE.format(name), server_uri)

                existing = keyring.get_password(self._SYSTEM, self._NAMES)

                if existing is None:
                    existing = [name]
                else:
                    existing = json.loads(existing)
                    if name not in existing:
                        existing.append(name)

                keyring.set_password(self._SYSTEM, self._NAMES, json.dumps(existing))

            else:
                open(self._REMOTE, "w+").write(server_uri)

        except Exception as error:
            raise error

    def get_remote(self, cluster):
        """
        Retrives the cluster URL

        :param cluster: Cluster name
        :type cluster: str
        :return:
        """

        if self.clusterSetting is ClusterSetting.MEMORY:

            return self.get_memory_cluster(cluster).cluster

        if self.clusterSetting is ClusterSetting.KEY_RING:

            if cluster is None:
                cluster = self.get_default()

            value = keyring.get_password(self._SYSTEM, self._REMOTE.format(cluster))
            if value is None:
                raise Exception(
                    "Remote server must be setup first by running, ah cluster add")

            return value

        try:
            return open(self._REMOTE, "r+").read()
        except Exception as err:
            raise err

    def remove_cluster(self, name):
        """
        Removes the given cluster

        :param name: Cluster name or ID
        :type name: str
        :return:
        """
        if self.clusterSetting is ClusterSetting.KEY_RING:

            keyring.delete_password(self._SYSTEM, self._USERNAME.format(name))
            keyring.delete_password(self._SYSTEM, self._PASSWORD.format(name))
            keyring.delete_password(self._SYSTEM, self._REMOTE.format(name))
            keyring.delete_password(self._SYSTEM, self._AUTHTOKEN.format(name))

            existing = keyring.get_password(self._SYSTEM, self._NAMES)

            if existing is not None:
                existing = json.loads(existing)
                existing.remove(name)
                keyring.set_password(self._SYSTEM, self._NAMES, json.dumps(existing))

            if name == self.get_default():
                keyring.delete_password(self._SYSTEM, self._DEFAULT)

    def set_username(self, username, cluster):
        """
        Overrides the username value for a given cluster

        :param username: New username Value
        :type username: str
        :param cluster: Cluster name to be retrieved
        :type cluster: str
        :return:
        """
        if self.clusterSetting is ClusterSetting.KEY_RING:
            keyring.set_password(self._SYSTEM, self._USERNAME.format(cluster), username)

    def get_username(self, cluster):
        """
        Retrieves the username of the given cluster

        :param cluster: Cluster name to be retrieved
        :type cluster: str
        :return:
        """
        if self.clusterSetting is ClusterSetting.MEMORY:

            return self.get_memory_cluster(cluster).userName

        if self.clusterSetting is ClusterSetting.KEY_RING:

            if cluster is None:
                cluster = self.get_default()

            value = keyring.get_password(self._SYSTEM, self._USERNAME.format(cluster))
            if value is None:
                raise Exception(
                    "Remote server must be setup first by running, ah cluster add")
            return value

        return None

    def get_password(self, cluster):
        """
        Retrieves the password of the given cluster

        :param cluster: Cluster name to be retrieved
        :type cluster: str
        :return:
        """
        if self.clusterSetting is ClusterSetting.MEMORY:

            return self.get_memory_cluster(cluster).password

        if self.clusterSetting is ClusterSetting.KEY_RING:

            if cluster is None:
                cluster = self.get_default()

            value = keyring.get_password(self._SYSTEM, self._PASSWORD.format(cluster))
            if value is None:
                raise Exception(
                    "Remote server must be setup first by running, ah cluster add")
            return value

        return None

    def set_password(self, password, cluster):
        """
        Overrides the password value for a given cluster

        :param password: New password Value
        :type password: str
        :param cluster: Cluster name to be retrieved
        :type cluster: str
        :return:
        """
        if self.clusterSetting is ClusterSetting.KEY_RING:

            keyring.set_password(self._SYSTEM, self._PASSWORD.format(cluster), password)

        else:
            pass

    def get_all_clusters(self):
        """
        Retrieves the list of all stored clusters

        :return:
        """
        if self.clusterSetting is ClusterSetting.MEMORY:

            return self.clusters.values()

        if self.clusterSetting is ClusterSetting.KEY_RING:

            names = keyring.get_password(self._SYSTEM, self._NAMES)
            if names is not None:
                existing = json.loads(names)
                clusters = []
                for item in existing:
                    clusters.append(
                        {
                            "name": item,
                            "url": keyring.get_password(self._SYSTEM, self._REMOTE.format(item)),
                            "username": keyring.get_password(
                                self._SYSTEM,
                                self._USERNAME.format(item)
                            ),
                            "default": item == self.get_default(),
                            "assetsync": True
                        }
                    )
                return clusters

            return []

        return []

    def set_auth_token(self, token, cluster):
        """
        Stores a new authentication token

        :param token: Auth token
        :type token: str
        :param cluster: Cluster name
        :type cluster: str
        :return:
        """

        if self.clusterSetting is ClusterSetting.MEMORY:

            self.clusters[cluster].token = token

        elif self.clusterSetting is ClusterSetting.KEY_RING:

            keyring.set_password(self._SYSTEM, self._AUTHTOKEN.format(cluster), token)
        else:

            open(self._AUTHTOKEN, "w+").write(token)

    def get_auth_token(self, cluster):
        """
        Retrieves the stored authentication token. This method does not make
        a login request.

        :param cluster: Cluster name
        :type cluster: str
        :return:
        """
        if self.clusterSetting is ClusterSetting.MEMORY:

            return self.get_memory_cluster(cluster).token

        if self.clusterSetting is ClusterSetting.KEY_RING:

            if cluster is None:
                cluster = self.get_default()

            value = keyring.get_password(self._SYSTEM, self._AUTHTOKEN.format(cluster))

            return value

        try:
            return open(self._AUTHTOKEN, "r+").read()
        except Exception as err:
            raise err




# Copyright 2021 Element Analytics, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from tests import *


class TestOrgAdminDefaultOrg(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.org_admin = OrgAdmin(
            cluster=cluster_name,
            props=props,
            org_id=test_org
        )

    def test_get_org_info(self):
        """
        Verify the sdk can retrieve org info

        :return:
        """

        info, _ = self.org_admin.get_org_info()

        self.assertTrue("id" in info, info)
        self.assertEqual(str(info["id"]), str(test_org), info)

    def test_submit_sensor_diagnostics(self):
        info = self.org_admin.submit_sensor_diagnostics(
            content=open("tests/sensor.json", "r+").read()
        )
        self.assertTrue("reportId" in info)

    def test_create_org(self):
        org_name = "zQA-{}".format(str(uuid.uuid4()))
        create = self.org_admin.create_organization(org_name=org_name)
        self.assertTrue(
            str(create).strip().isdigit(),
            "Org Id is created and its a number"
        )

    @unittest.skip("Dont deletr orgs")
    def test_delete_org(self):
        org_name = "zQA-{}".format(str(uuid.uuid4()))
        create = self.org_admin.create_organization(org_name=org_name)

        org_admin_2 = OrgAdmin(
            cluster=cluster_name,
            props=props,
            org_id=create
        )

        delete = org_admin_2.delete_organization()

        self.assertTrue("message" in delete)

    @unittest.skip("Random passwords are hard")
    def test_invite_machine_user(self):
        invite = self.org_admin.invite_machine_user(
            id=str(uuid.uuid4()),
            password='xY+psKZ9c]5/N%,26{}'.format(str(uuid.uuid4())),
            fullname=str(uuid.uuid4()),
            role="Admin"
        )

        self.assertTrue("userId" in invite)

    @unittest.skip("Rename must be broken")
    def test_rename_organization(self):
        org_name = "zQA-{}".format(str(uuid.uuid4()))
        create = self.org_admin.create_organization(org_name=org_name)

        org_admin_2 = OrgAdmin(
            cluster=cluster_name,
            props=props,
            org_id=create
        )

        rename = org_admin_2.rename_organization(new_name="zQA-{}".format(str(uuid.uuid4())))

        self.assertEqual(str(rename["id"]), str(create), rename)
        self.assertTrue("name" in rename)
